"""Honkai battle chronicle models."""

from __future__ import annotations

import datetime
import re
import typing

import pydantic

from genshin.models.honkai import battlesuit
from genshin.models.model import Aliased, APIModel, Unique

__all__ = ["ELF", "Boss", "ElysianRealm", "MemorialArena", "MemorialBattle", "OldAbyss", "SuperstringAbyss"]

REMEMBRANCE_SIGILS: dict[int, tuple[str, int]] = {
    119301: ("The MOTH Insignia", 1),
    119302: ("Home Lost", 1),
    119303: ("False Hope", 1),
    119304: ("Tin Flask", 1),
    119305: ("Ruined Legacy", 1),
    119306: ("Burden", 2),
    119307: ("Gold Goblet", 2),
    119308: ("Mad King's Mask", 2),
    119309: ("Light as a Bodhi Leaf", 2),
    119310: ("Forget-Me-Not", 2),
    119311: ("Forbidden Seed", 2),
    119312: ("Memory", 2),
    119313: ("Crystal Rose", 2),
    119314: ("Abandoned", 3),
    119315: ("Good Old Days", 3),
    119316: ("Shattered Shackles", 3),
    119317: ("Heavy as a Million Lives", 3),
    119318: ("Stained Sakura", 3),
    119319: ("The First Scale", 3),
    119320: ("Resolve", 3),
    119321: ("Thorny Crown", 3),
    119326: ("Veil of Tears", 2),
    119327: ("Pseudo Miracle", 2),
    119328: ("Fragile Friend", 2),
    119329: ("Rainbow of Absence", 2),
    119330: ("Feast of Emptiness", 2),
    119331: ("It Will Be Written", 4),
    119332: ("Because of You", 4),
    119333: ("Boundless Feeling", 4),
    119334: ("Dreamful Gold", 4),
    119335: ("Falling in Past Light", 4),
    119336: ("An Old Pal's Legacy", 4),
    119337: ("Empty Like Shala", 4),
    119338: ("Tsukimi Himiko", 4),
    119339: ("Out of Reach", 4),
    119340: ("Boundless Logos", 4),
    119341: ("The Lonely Moon", 4),
    119342: ("Hometown", 4),
    119343: ("Awakening", 4),
    119344: ("Proof of Good and Evil", 3),
    119345: ("Faraway Ship", 3),
    119346: ("Ravenous Gully", 3),
    119347: ("Grey-scale Rainbow", 3),
    119348: ("Nine Lives", 3),
    119349: ("Key to the Deep", 1),
}


# GENERIC


class Boss(APIModel, Unique):
    """Represents a Boss encountered in Abyss or Memorial Arena."""

    id: int
    name: str
    icon: str = Aliased("avatar")

    @pydantic.field_validator("icon")
    def __fix_url(cls, url: str) -> str:
        # I noticed that sometimes the urls are returned incorrectly, which appears to be
        # a problem on the hoyolab website too, so I expect this to be fixed sometime.
        # For now, this hotfix seems to work.
        return re.sub(r"/boss_\d+\.", lambda m: str.upper(m[0]), url, flags=re.IGNORECASE)


class ELF(APIModel, Unique):
    """Represents an ELF equipped for a battle."""

    id: int
    name: str
    icon: str = Aliased("avatar")
    rarity: str
    upgrade_level: int = Aliased("star")

    @pydantic.field_validator("rarity", mode="before")
    def __fix_rank(cls, rarity: typing.Union[int, str]) -> str:
        if isinstance(rarity, str):
            return rarity

        # ELFs come in rarities A and S, API returns 4 and 5, respectively
        return ["A", "S"][rarity - 4]


# ABYSS


def get_abyss_rank_mi18n(rank: int, tier: int) -> str:
    """Turn the rank returned by the API into the respective rank name displayed in-game."""
    if tier == 4:
        mod = ("1", "2_1", "2_2", "2_3", "3_1", "3_2", "3_3", "4", "5")[rank - 1]
    else:
        mod = str(rank)
    return f"bbs/level{mod}"


class BaseAbyss(APIModel):
    """Represents one cycle of abyss.

    (3 days per cycle, 2 cycles per week)
    """

    # somewhat feel like this is overkill

    raw_tier: int = Aliased("area")
    score: int
    lineup: typing.Sequence[battlesuit.Battlesuit]
    boss: Boss
    elf: typing.Optional[ELF]


class OldAbyss(BaseAbyss):
    """Represents once cycle of Quantum Singularis or Dirac Sea.

    Exclusive to players of level 80 and below.
    """

    end_time: datetime.datetime = Aliased("time_second")
    raw_type: str = Aliased("type")
    result: str = Aliased("reward_type")
    raw_rank: int = Aliased("level")

    @pydantic.field_validator("raw_rank", mode="before")
    def __normalize_level(cls, rank: str) -> int:
        # The latestOldAbyssReport endpoint returns ranks as D/C/B/A,
        # while newAbyssReport returns them as 1/2/3/4(/5) respectively.
        # Having them as ints at base seems more useful than strs.
        # (in-game, they use the same names (Sinful, Agony, etc.))
        if isinstance(rank, int):
            return rank

        return 69 - ord(rank)


class SuperstringAbyss(BaseAbyss):
    """Represents one cycle of Superstring Abyss, exclusive to players of level 81 and up."""

    # NOTE endpoint: game_record/honkai3rd/api/latestOldAbyssReport

    end_time: datetime.datetime = Aliased("updated_time_second")
    raw_tier: int = 4  # Not returned by API, always the case
    placement: int = Aliased("rank")
    trophies_gained: int = Aliased("settled_cup_number")
    end_trophies: int = Aliased("cup_number")
    raw_start_rank: int = Aliased("level")
    raw_end_rank: int = Aliased("settled_level")

    @property
    def start_trophies(self) -> int:
        return self.end_trophies - self.trophies_gained


# MEMORIAL ARENA


def prettify_MA_rank(rank: int) -> str:  # Independent of mi18n
    """Turn the rank returned by the API into the respective rank name displayed in-game."""
    brackets = (0, 0.20, 2, 7, 17, 35, 65, 100)
    return f"{brackets[rank - 1]:1.2f} ~ {brackets[rank]:1.2f}"


class MemorialBattle(APIModel):
    """Represents weekly performance against a single Memorial Arena boss."""

    score: int
    lineup: typing.Sequence[battlesuit.Battlesuit]
    elf: typing.Optional[ELF]
    boss: Boss


class MemorialArena(APIModel):
    """Represents aggregate weekly performance for the entire Memorial Arena rotation."""

    score: int
    ranking: float = Aliased("ranking_percentage")
    raw_rank: int = Aliased("rank")
    raw_tier: int = Aliased("area")
    end_time: datetime.datetime = Aliased("time_second")
    battle_data: typing.Sequence[MemorialBattle] = Aliased("battle_infos")

    @property
    def rank(self) -> str:
        """The user's Memorial Arena rank as displayed in-game."""
        return prettify_MA_rank(self.raw_rank)


# ELYSIAN REALMS
# TODO: Implement a way to link response_json["avatar_transcript"] data to be added to
#       ER lineup data; will require new Battlesuit subclass.


class Condition(APIModel):
    """Represents a debuff picked at the beginning of an Elysian Realms run."""

    name: str
    description: str = Aliased("desc")
    difficulty: int


class Signet(APIModel):
    """Represents a buff Signet picked in an Elysian Realms run."""

    id: int
    icon: str
    number: int

    @property
    def name(self) -> str:
        return [
            "Deliverance",
            "Gold",
            "Decimation",
            "Bodhi",
            "Setsuna",
            "Infinity",
            "Vicissitude",
            "Ego",
            "Unknown",      # Unused, no signet with id 9.
            "Discipline",   # icon 12
            "Helix",        # icon 13
            "Daybreak",     # icon 9
            "Stars",        # icon 10
            "Reverie",      # icon 11
        ][self.id - 1]  # fmt: skip

    def get_scaled_icon(self, scale: typing.Literal[1, 2, 3] = 2) -> str:
        if not 1 <= scale <= 3:
            raise ValueError("Scale must lie between 1 and 3.")

        return self.icon.replace("@2x", "" if scale == 1 else f"@{scale}x")


class RemembranceSigil(APIModel):
    """Represents a Remembrance Sigil from Elysian Realms."""

    icon: str

    @property
    def id(self) -> int:
        match = re.match(r".*/(\d+).png", self.icon)
        return int(match[1]) if match else 0

    @property
    def name(self) -> str:
        sigil = REMEMBRANCE_SIGILS.get(self.id)
        return sigil[0] if sigil else "Unknown"

    @property
    def rarity(self) -> int:
        sigil = REMEMBRANCE_SIGILS.get(self.id)
        return sigil[1] if sigil else 1


class ElysianRealm(APIModel):
    """Represents one completed run of Elysean Realms."""

    completed_at: datetime.datetime = Aliased("settle_time_second")
    floors_cleared: int = Aliased("level")
    score: int
    difficulty: int = Aliased("punish_level")
    conditions: typing.Sequence[Condition]
    signets: typing.Sequence[Signet] = Aliased("buffs")
    leader: battlesuit.Battlesuit = Aliased("main_avatar")
    supports: typing.Sequence[battlesuit.Battlesuit] = Aliased("support_avatars")
    elf: typing.Optional[ELF]
    remembrance_sigil: RemembranceSigil = Aliased("extra_item_icon")

    @pydantic.field_validator("remembrance_sigil", mode="before")
    def __extend_sigil(cls, sigil: typing.Any) -> typing.Any:
        if isinstance(sigil, str):
            return dict(icon=sigil)

        return sigil

    @property
    def lineup(self) -> typing.Sequence[battlesuit.Battlesuit]:
        return [self.leader, *self.supports]



"""Genshin calculator models."""

from __future__ import annotations

import collections
import enum
import typing

import pydantic

from genshin.models.model import Aliased, APIModel, Unique

from . import character

__all__ = [
    "CALCULATOR_ARTIFACTS",
    "CALCULATOR_ELEMENTS",
    "CALCULATOR_WEAPON_TYPES",
    "CalculatorArtifact",
    "CalculatorArtifactResult",
    "CalculatorCharacter",
    "CalculatorCharacterDetails",
    "CalculatorConsumable",
    "CalculatorFurnishing",
    "CalculatorFurnishingResults",
    "CalculatorResult",
    "CalculatorTalent",
    "CalculatorWeapon",
]

CALCULATOR_ELEMENTS: typing.Mapping[int, str] = {
    1: "Pyro",
    2: "Anemo",
    3: "Geo",
    4: "Dendro",
    5: "Electro",
    6: "Hydro",
    7: "Cryo",
}
CALCULATOR_WEAPON_TYPES: typing.Mapping[int, str] = {
    0: "Unknown",
    1: "Sword",
    10: "Catalyst",
    11: "Claymore",
    12: "Bow",
    13: "Polearm",
}
CALCULATOR_ARTIFACTS: typing.Mapping[int, str] = {
    1: "Flower of Life",
    2: "Plume of Death",
    3: "Sands of Eon",
    4: "Goblet of Eonothem",
    5: "Circlet of Logos",
}


class MaterialSource(str, enum.Enum):
    """Source of the material."""

    SPECIAL = "MT_SPECIAL"
    MONSTER = "MT_MONSTER"
    WORLD = "MT_WORLD"
    DOMAIN = "MT_DUNGEON"
    WEEKLY_BOSS = "MT_WEEKLY_DUNGEON"


class PartialCalculatorCharacter(APIModel):
    id: int
    icon: str
    rarity: int = Aliased("avatar_level")


class PartialCalculatorWeapon(APIModel):
    id: int
    icon: str
    rarity: int = Aliased("weapon_level")


class CalculatorCharacter(character.BaseCharacter):
    """Character meant to be used with calculators."""

    element: str = Aliased("element_attr_id")
    weapon_type: str = Aliased("weapon_cat_id")
    level: int = Aliased("level_current", default=0)
    max_level: int

    @pydantic.field_validator("element", mode="before")
    def __parse_element(cls, v: typing.Any) -> str:
        if isinstance(v, str):
            return v

        return CALCULATOR_ELEMENTS[int(v)]

    @pydantic.field_validator("weapon_type", mode="before")
    def __parse_weapon_type(cls, v: typing.Any) -> str:
        if isinstance(v, str):
            return v

        return CALCULATOR_WEAPON_TYPES[int(v)]


class CalculatorWeapon(APIModel, Unique):
    """Weapon meant to be used with calculators."""

    id: int
    name: str
    icon: str
    rarity: int = Aliased("weapon_level")
    type: str = Aliased("weapon_cat_id")
    level: int = Aliased("level_current", default=0)
    max_level: int

    @pydantic.field_validator("type", mode="before")
    def __parse_weapon_type(cls, v: typing.Any) -> str:
        if isinstance(v, str):
            return v

        return CALCULATOR_WEAPON_TYPES[int(v)]


class CalculatorArtifact(APIModel, Unique):
    """Artifact meant to be used with calculators."""

    id: int
    name: str
    icon: str
    rarity: int = Aliased("reliquary_level")
    pos: int = Aliased("reliquary_cat_id")
    level: int = Aliased("level_current", default=0)
    max_level: int

    @property
    def pos_name(self) -> str:
        return CALCULATOR_ARTIFACTS[self.pos]


class CalculatorTalent(APIModel, Unique):
    """Talent of a character meant to be used with calculators."""

    id: int
    group_id: int  # proudSkillGroupId
    name: str
    icon: str
    level: int = Aliased("level_current", default=0)
    max_level: int

    @property
    def type(self) -> typing.Literal["attack", "skill", "burst", "passive", "dash"] | None:
        """The type of the talent, parsed from the group id.

        Does not work for traveler!
        """
        # special cases
        if self.id == self.group_id:
            return "passive"  # maybe hoyo does this for unapgradables?

        if len(str(self.id)) == 6:  # in candSkillDepotIds
            return "attack"

        # 4139 -> group=41 identifier=3 order=9
        _, relevant = divmod(self.group_id, 100)
        identifier, order = divmod(relevant, 10)

        if identifier == 2:
            return "passive"
        elif order == 1:
            return "attack"
        elif order == 2:
            return "skill"
        elif order == 9:
            return "burst"
        elif order == 3:
            return "dash"
        else:
            return None

    @property
    def upgradeable(self) -> bool:
        """Whether this talent can be leveled up."""
        return self.type not in ("passive", "dash")

    def __int__(self) -> int:
        return self.group_id


class CalculatorFurnishing(APIModel, Unique):
    """Furnishing meant to be used with calculators."""

    id: int
    name: str
    icon: str = Aliased("icon_url")
    rarity: int = Aliased("level")

    amount: typing.Optional[int] = Aliased("num", default=None)


class CalculatorCharacterDetails(APIModel):
    """Details of a synced calculator character."""

    weapon: CalculatorWeapon = Aliased("weapon")
    talents: typing.Sequence[CalculatorTalent] = Aliased("skill_list")
    artifacts: typing.Sequence[CalculatorArtifact] = Aliased("reliquary_list")

    @pydantic.field_validator("talents")
    def __correct_talent_current_level(cls, v: typing.Sequence[CalculatorTalent]) -> typing.Sequence[CalculatorTalent]:
        # passive talent have current levels at 0 for some reason
        talents: list[CalculatorTalent] = []

        for talent in v:
            if talent.max_level == 1 and talent.level == 0:
                raw = talent.model_dump()
                raw["level"] = 1
                talent = CalculatorTalent(**raw)

            talents.append(talent)

        return v

    @property
    def upgradeable_talents(self) -> typing.Sequence[CalculatorTalent]:
        """All talents that can be leveled up."""
        if self.talents[2].type == "dash":
            return (self.talents[0], self.talents[1], self.talents[3])
        else:
            return (self.talents[0], self.talents[1], self.talents[2])


class CalculatorConsumable(APIModel, Unique):
    """Item consumed when upgrading."""

    id: int
    name: str
    icon: str
    rarity: int = Aliased("level")
    amount: int = Aliased("num", default=1)
    lacking: int = Aliased("lack_num", default=0)
    wiki_url: str


class CalculatorArtifactResult(APIModel):
    """Calculation result for a specific artifact."""

    artifact_id: int = Aliased("reliquary_id")
    materials: typing.Sequence[CalculatorConsumable] = Aliased("id_consume_list")


class CalculatorTalentInfo(APIModel):
    """Information about a talent in the calculator."""

    talent_id: int = Aliased("id")
    current: int = Aliased("level_current", default=0)
    target: int = Aliased("level_target", default=0)


class CalculatorTalentResult(APIModel):
    """Calculation result for a specific talent."""

    info: CalculatorTalentInfo = Aliased("skill_info")
    materials: typing.Sequence[CalculatorConsumable] = Aliased("consume_list")


class CalculatorMonster(APIModel):
    """Monster that drops a material."""

    id: int = Aliased("monster_id")
    name: str = Aliased("monster_name")
    icon: str = Aliased("monster_icon")
    map_url: str = Aliased("monster_map_url")


class CalculatorDomain(APIModel):
    """Domain that drops a material."""

    name: str = Aliased("dungeon_name")
    weekdays: typing.Sequence[int] = Aliased("drop_day")
    """List of weekdays the domain is open, 0=Monday, 6=Sunday."""
    calendar_url: str = Aliased("calendar_link")


class RemainingMaterialDetail(APIModel):
    """Details of remaining required materials for a character."""

    materials: typing.Sequence[CalculatorConsumable] = Aliased("consume")

    characters: typing.Sequence[PartialCalculatorCharacter] = Aliased("avatars")
    """Characters these materials are needed for, empty if accessing weapons."""
    weapons: typing.Sequence[PartialCalculatorWeapon] = Aliased("weapons")
    """Weapons these materials are needed for, empty if accessing characters or talents."""

    material_source: MaterialSource

    monster: typing.Optional[CalculatorMonster] = None
    """Monster info, present if material_source is MT_MONSTER or MT_WEEKLY_DUNGEON."""
    map_url: typing.Optional[str] = None
    """URL to interactive map, present if material_source is MT_WORLD."""
    domain: typing.Optional[CalculatorDomain] = Aliased("dungeon_calendar", default=None)
    """Domain info, may be present if material_source is MT_DUNGEON."""


class RemainingMaterialDetails(APIModel):
    """Details of remaining required materials for a character."""

    characters: typing.Sequence[RemainingMaterialDetail] = Aliased("avatar_consume")
    weapon: typing.Sequence[RemainingMaterialDetail] = Aliased("weapon_consume")
    talents: typing.Sequence[RemainingMaterialDetail] = Aliased("avatar_skill_consume")


class CalculatorResult(APIModel):
    """Calculation result."""

    character: list[CalculatorConsumable] = Aliased("avatar_consume")
    weapon: list[CalculatorConsumable] = Aliased("weapon_consume")
    talents: list[CalculatorTalentResult] = Aliased("skills_consume")
    artifacts: list[CalculatorArtifactResult] = Aliased("reliquary_consume")

    lineup_recommendation: typing.Optional[str] = Aliased("lineup_recommend", default=None)

    @property
    def total(self) -> typing.Sequence[CalculatorConsumable]:
        talents = [i for t in self.talents for i in t.materials]
        artifacts = [i for a in self.artifacts for i in a.materials]
        combined = self.character + self.weapon + talents + artifacts

        grouped: dict[int, list[CalculatorConsumable]] = collections.defaultdict(list)
        for i in combined:
            grouped[i.id].append(i)

        total = [
            CalculatorConsumable(
                id=x[0].id,
                name=x[0].name,
                icon=x[0].icon,
                wiki_url=x[0].wiki_url,
                rarity=x[0].rarity,
                lacking=x[0].lacking,
                amount=sum(i.amount for i in x),
            )
            for x in grouped.values()
        ]

        return total


class CalculatorBatchResult(APIModel):
    """Batch calculation result."""

    characters: list[CalculatorResult] = Aliased("items")
    available_materials: list[CalculatorConsumable] = Aliased("available_material")
    total_materials: list[CalculatorConsumable] = Aliased("overall_consume")
    total_material_details: RemainingMaterialDetails = Aliased("overall_material_consume")
    individual_results: typing.Sequence[CalculatorBatchResult] = Aliased("single_role_result")

    has_user_info: bool

    @property
    def total_remaining_required_materials(self) -> typing.Sequence[CalculatorConsumable]:
        return [m for m in self.total_materials if m.lacking > 0]


class CalculatorFurnishingResults(APIModel):
    """Furnishing calculation result."""

    furnishings: list[CalculatorConsumable] = Aliased("list")

    @property
    def total(self) -> typing.Sequence[CalculatorConsumable]:
        return self.furnishings



"""Modified pydantic model."""

from __future__ import annotations

import abc
import datetime
import enum
import logging
import typing
from typing import Annotated

import pydantic

from genshin.constants import CN_TIMEZONE

__all__ = ["APIModel", "Aliased", "Unique"]

logger: logging.Logger = logging.getLogger(__name__)


class APIModel(pydantic.BaseModel):
    """Modified pydantic model."""

    model_config: pydantic.ConfigDict = pydantic.ConfigDict(arbitrary_types_allowed=True)  # type: ignore


class Unique(abc.ABC):
    """A hashable model with an id."""

    id: int

    def __int__(self) -> int:
        return hash(self.id)

    def __hash__(self) -> int:
        return hash(self.id)


def Aliased(
    alias: typing.Optional[str] = None,
    default: typing.Any = ...,
    **kwargs: typing.Any,
) -> typing.Any:
    """Create an aliased field."""
    return pydantic.Field(default, alias=alias, **kwargs)


def add_timezone(value: datetime.datetime) -> datetime.datetime:
    return value.replace(tzinfo=CN_TIMEZONE)


def convert_datetime(value: typing.Optional[typing.Mapping[str, typing.Any]]) -> datetime.datetime:
    if value:
        value = typing.cast(typing.Dict[str, typing.Any], value)
        tzinfo = value.pop("tzinfo", None)
        if tzinfo is not None:
            tz = datetime.timezone(datetime.timedelta(hours=tzinfo))
        else:
            tz = None

        return datetime.datetime(tzinfo=tz, **value)

    msg = f"Invalid datetime value provided: {value!r}"
    raise ValueError(msg)


def parse_timestamp(value: typing.Optional[typing.Union[str, int]]) -> datetime.datetime:
    """Parse a timestamp into a datetime object."""
    if isinstance(value, str):
        value = int(value)
    if isinstance(value, int):
        return datetime.datetime.fromtimestamp(value).replace(tzinfo=datetime.timezone.utc)
    raise ValueError(f"Invalid timestamp value provided: {value!r}")


InputValue = typing.TypeVar("InputValue", str, int, enum.Enum)
EnumType = typing.TypeVar("EnumType", bound=enum.Enum)


def prevent_enum_error(value: InputValue, cls: typing.Type[EnumType]) -> typing.Union[EnumType, InputValue]:
    """Prevent enum error by returning the value as is."""
    try:
        return cls(value)
    except ValueError:
        logger.warning("Unknown %s: %r, please report to developer", cls.__name__, value)
        return value


TZDateTime = Annotated[datetime.datetime, pydantic.AfterValidator(add_timezone)]
DateTime = Annotated[datetime.datetime, pydantic.BeforeValidator(convert_datetime)]
UnixDateTime = Annotated[datetime.datetime, pydantic.BeforeValidator(parse_timestamp)]
LevelField = Annotated[int, pydantic.Field(ge=1)]



import datetime
import enum
import typing

import pydantic

from genshin.models.model import Aliased, APIModel, DateTime, TZDateTime
from genshin.models.zzz.character import ZZZElementType

from .common import ChallengeBangboo

__all__ = (
    "ChallengeBangboo",
    "ShiyuDefense",
    "ShiyuDefenseBangboo",
    "ShiyuDefenseBuff",
    "ShiyuDefenseCharacter",
    "ShiyuDefenseFloor",
    "ShiyuDefenseMonster",
    "ShiyuDefenseNode",
    "ShiyuDefenseV1",
    "ShiyuDefenseV2",
    "ShiyuMonsterElementEffect",
    "ShiyuMonsterElementEffects",
    "ShiyuV2FifthFloor",
    "ShiyuV2FifthFloorLayer",
    "ShiyuV2FourthFloor",
    "ShiyuV2FourthFloorLayer",
)


class ShiyuMonsterElementEffect(enum.IntEnum):
    """Shiyu Defense monster element effect enum."""

    WEAKNESS = 1
    RESISTANCE = -1
    NEUTRAL = 0


class ShiyuDefenseBangboo(ChallengeBangboo):
    """Shiyu Defense bangboo model."""


class ShiyuDefenseCharacter(APIModel):
    """Shiyu Defense character model."""

    id: int
    level: int
    rarity: typing.Literal["S", "A"]
    element: ZZZElementType = Aliased("element_type")
    icon: str = Aliased("role_square_url")
    mindscape: int = Aliased("rank")


class ShiyuDefenseBuff(APIModel):
    """Shiyu Defense buff model."""

    name: str = Aliased("title")
    description: str = Aliased("text")


class ShiyuMonsterElementEffects(pydantic.BaseModel):
    """Shiyu Defense monster element effects model."""

    ice: ShiyuMonsterElementEffect = Aliased("ice_weakness")
    fire: ShiyuMonsterElementEffect = Aliased("fire_weakness")
    electric: ShiyuMonsterElementEffect = Aliased("elec_weakness")
    ether: ShiyuMonsterElementEffect = Aliased("ether_weakness")
    physical: ShiyuMonsterElementEffect = Aliased("physics_weakness")


class ShiyuDefenseMonster(APIModel):
    """Shiyu Defense monster model."""

    id: int
    name: str
    weakness: typing.Union[ZZZElementType, int] = Aliased(
        "weak_element_type", deprecated="ShiyuDefenseMonster.weakness doesn't return anything meaningful anymore."
    )
    level: int
    element_effects: ShiyuMonsterElementEffects

    @pydantic.model_validator(mode="before")
    @classmethod
    def __nest_element_effects(cls, v: dict[str, typing.Any]) -> dict[str, typing.Any]:
        v["element_effects"] = {
            "ice_weakness": v["ice_weakness"],
            "fire_weakness": v["fire_weakness"],
            "elec_weakness": v["elec_weakness"],
            "ether_weakness": v["ether_weakness"],
            "physics_weakness": v["physics_weakness"],
        }
        return v


class ShiyuDefenseNode(APIModel):
    """Shiyu Defense node model."""

    characters: list[ShiyuDefenseCharacter] = Aliased("avatars")
    bangboo: typing.Optional[ShiyuDefenseBangboo] = Aliased("buddy", default=None)
    recommended_elements: list[ZZZElementType] = Aliased("element_type_list")
    enemies: list[ShiyuDefenseMonster] = Aliased("monster_info")
    battle_time: typing.Optional[datetime.timedelta] = None

    @pydantic.field_validator("enemies", mode="before")
    @classmethod
    def __convert_enemies(cls, value: dict[typing.Literal["level", "list"], typing.Any]) -> list[ShiyuDefenseMonster]:
        level = value["level"]
        result: list[ShiyuDefenseMonster] = []
        for monster in value["list"]:
            monster["level"] = level
            result.append(ShiyuDefenseMonster(**monster))
        return result


class ShiyuDefenseFloor(APIModel):
    """Shiyu Defense floor model."""

    index: int = Aliased("layer_index")
    rating: typing.Literal["S", "A", "B"]
    id: int = Aliased("layer_id")
    buffs: list[ShiyuDefenseBuff]
    node_1: ShiyuDefenseNode
    node_2: ShiyuDefenseNode
    challenge_time: TZDateTime = Aliased("floor_challenge_time")
    name: str = Aliased("zone_name")

    @pydantic.field_validator("challenge_time", mode="before")
    @classmethod
    def __parse_datetime(cls, value: typing.Mapping[str, typing.Any]) -> typing.Optional[TZDateTime]:
        if value:
            return datetime.datetime(**value)
        return None


class ShiyuDefenseV1(APIModel):
    """ZZZ Shiyu Defense V1 model."""

    schedule_id: int
    begin_time: typing.Optional[DateTime] = Aliased("hadal_begin_time")
    end_time: typing.Optional[DateTime] = Aliased("hadal_end_time")
    has_data: bool
    ratings: typing.Mapping[typing.Literal["S", "A", "B"], int] = Aliased("rating_list")
    floors: list[ShiyuDefenseFloor] = Aliased("all_floor_detail")
    fastest_clear_time: int = Aliased("fast_layer_time")
    """Fastest clear time this season in seconds."""
    max_floor: int = Aliased("max_layer")

    @pydantic.field_validator("ratings", mode="before")
    @classmethod
    def __convert_ratings(
        cls, v: list[dict[typing.Literal["times", "rating"], typing.Any]]
    ) -> typing.Mapping[typing.Literal["S", "A", "B"], int]:
        return {d["rating"]: d["times"] for d in v}

    @pydantic.computed_field  # type: ignore[prop-decorator]
    @property
    def total_clear_time(self) -> int:
        """Total clear time for all floors in seconds."""
        total = 0
        for floor in self.floors:
            for node in (floor.node_1, floor.node_2):
                if node.battle_time is None:
                    continue
                total += int(node.battle_time.total_seconds())
        return total


ShiyuDefense = ShiyuDefenseV1  # Backward compatibility


class ShiyuV2FifthFloorLayer(APIModel):
    """ZZZ Shiyu Defense V2 fifth floor layer model."""

    id: int = Aliased("layer_id")
    rating: typing.Literal["S", "A", "B"]
    buff: ShiyuDefenseBuff = Aliased("buffer")
    score: int
    max_score: int
    clear_time: int = Aliased("battle_time")
    """Clear time in seconds."""
    boss_icon: str = Aliased("monster_pic")
    bangboo: typing.Optional[ShiyuDefenseBangboo] = Aliased("buddy", default=None)
    characters: typing.Sequence[ShiyuDefenseCharacter] = Aliased("avatar_list")


class ShiyuV2FifthFloor(APIModel):
    """ZZZ Shiyu Defense V2 fifth floor model."""

    layers: typing.Sequence[ShiyuV2FifthFloorLayer] = Aliased("layer_challenge_info_list")


class ShiyuV2FourthFloorLayer(APIModel):
    """ZZZ Shiyu Defense V2 fourth floor layer model."""

    id: int = Aliased("layer_id")
    clear_time: int = Aliased("battle_time")
    """Clear time in seconds."""
    characters: typing.Sequence[ShiyuDefenseCharacter] = Aliased("avatar_list")
    bangboo: typing.Optional[ShiyuDefenseBangboo] = Aliased("buddy", default=None)


class ShiyuV2FourthFloor(APIModel):
    """ZZZ Shiyu Defense V2 fourth floor model."""

    buff: ShiyuDefenseBuff = Aliased("buffer")
    challenge_time: DateTime
    rating: typing.Literal["S", "A", "B"]
    layers: typing.Sequence[ShiyuV2FourthFloorLayer] = Aliased("layer_challenge_info_list")


class ShiyuV2BriefInfo(APIModel):
    """ZZZ Shiyu Defense V2 brief info."""

    score: int
    max_score: int
    rank_percent: str
    total_clear_time: int = Aliased("battle_time")
    rating: typing.Optional[typing.Literal["S+", "S", "A", "B"]]
    challenge_time: typing.Optional[DateTime] = None

    @pydantic.field_validator("rating", mode="before")
    def __parse_rating(
        cls, value: typing.Literal["S+", "S", "A", "B", ""]
    ) -> typing.Optional[typing.Literal["S+", "S", "A", "B"]]:
        return value or None

    @pydantic.field_validator("rank_percent", mode="before")
    def __parse_rank_percent(cls, value: int) -> str:
        return f"{value / 100}%"


class ShiyuDefenseV2(APIModel):
    """ZZZ Shiyu Defense V2 model."""

    schedule_id: int = Aliased("zone_id")
    begin_time: typing.Optional[DateTime] = Aliased("hadal_begin_time")
    end_time: typing.Optional[DateTime] = Aliased("hadal_end_time")
    passed_fifth_floor: bool = Aliased("pass_fifth_floor")

    brief_info: typing.Optional[ShiyuV2BriefInfo] = Aliased("brief", default=None)
    fourth_frontier: typing.Optional[ShiyuV2FourthFloor] = Aliased("fourth_layer_detail", default=None)
    fifth_frontier: typing.Optional[ShiyuV2FifthFloor] = Aliased("fitfh_layer_detail", default=None)  # Nice typo hoyo

    player_nickname: str = Aliased("nick_name")
    player_avatar: str = Aliased("icon")



"""Base battle chronicle component."""

import asyncio
import dataclasses
import typing
import warnings

from genshin import errors, models, types, utility
from genshin.client import cache, routes
from genshin.client.components import base
from genshin.client.manager import managers
from genshin.constants import GAME_LANGS
from genshin.models import hoyolab as hoyolab_models
from genshin.utility import deprecation

__all__ = ["BaseBattleChronicleClient"]


@dataclasses.dataclass(unsafe_hash=True)
class HoyolabCacheKey(cache.CacheKey):
    endpoint: str
    hoyolab_id: int
    lang: str


@dataclasses.dataclass(unsafe_hash=True)
class ChronicleCacheKey(cache.CacheKey):
    def __str__(self) -> str:
        return "chronicle" + ":" + super().__str__()

    game: types.Game
    endpoint: str
    uid: int
    lang: str
    params: tuple[typing.Any, ...] = ()


class BaseBattleChronicleClient(base.BaseClient):
    """Base battle chronicle component."""

    async def request_game_record(
        self,
        endpoint: str,
        *,
        lang: typing.Optional[str] = None,
        region: typing.Optional[types.Region] = None,
        game: typing.Optional[types.Game] = None,
        custom_route: typing.Optional[typing.Union[routes.Route, routes.InternationalRoute]] = None,
        **kwargs: typing.Any,
    ) -> typing.Mapping[str, typing.Any]:
        """Make a request towards the game record endpoint."""
        game = game or self.default_game

        if isinstance(custom_route, routes.InternationalRoute):
            base_url = routes.CARD_WAPI_URL.get_url(region or self.region)
        elif isinstance(custom_route, routes.Route):
            base_url = custom_route.get_url()
        else:
            if game is None:
                raise RuntimeError("No default game set.")
            base_url = routes.RECORD_URL.get_url(region or self.region, game)

        if game is not None and (lang or self.lang) not in GAME_LANGS[game]:
            warnings.warn(f"Language '{lang or self.lang}' is not supported for {game.name}. Defaulting to 'en-us'.")
            lang = "en-us"

        url = base_url / endpoint

        update_task = asyncio.create_task(utility.update_characters_any(lang or self.lang, lenient=True))

        data = await self.request_hoyolab(url, lang=lang, region=region, **kwargs)

        try:
            await update_task
        except Exception as e:
            warnings.warn(f"Failed to update characters: {e!r}")

        return data

    async def get_record_cards(
        self, hoyolab_id: typing.Optional[int] = None, *, lang: typing.Optional[str] = None
    ) -> list[models.hoyolab.RecordCard]:
        """Get a user's record cards."""
        hoyolab_id = hoyolab_id or self._get_hoyolab_id()

        cache_key = cache.cache_key("records", hoyolab_id=hoyolab_id, lang=lang or self.lang)
        if not (data := await self.cache.get(cache_key)):
            data = await self.request_game_record(
                "getGameRecordCard", lang=lang, params=dict(uid=hoyolab_id), custom_route=routes.CARD_WAPI_URL
            )

            if data["list"]:
                await self.cache.set(cache_key, data)
            else:
                raise errors.DataNotPublic({"retcode": 10102})

        return [models.hoyolab.RecordCard(**card) for card in data["list"]]

    @deprecation.deprecated("get_record_cards")
    async def get_record_card(
        self, hoyolab_id: typing.Optional[int] = None, *, lang: typing.Optional[str] = None
    ) -> models.hoyolab.RecordCard:
        """Get a user's record card."""
        cards = await self.get_record_cards(hoyolab_id, lang=lang)
        return cards[0]

    @managers.no_multi
    async def update_settings(
        self,
        setting: types.IDOr[hoyolab_models.RecordCardSetting],
        on: bool,
        *,
        game: typing.Optional[types.Game] = None,
    ) -> None:
        """Update user settings.

        Setting IDs:
            1: Show your Battle Chronicle on your profile.
            2: Show your Character Details in the Battle Chronicle.
            3: Enable your Real-Time Notes. (only for Genshin Impact)
        """
        if game is None and int(setting) == 3:
            game = types.Game.GENSHIN

        if game is None:
            if self.default_game is None:
                raise RuntimeError("No default game set.")

            game = self.default_game

        game_id = {types.Game.HONKAI: 1, types.Game.GENSHIN: 2, types.Game.STARRAIL: 6, types.Game.ZZZ: 8}[game]

        await self.request_game_record(
            "changeDataSwitch",
            method="POST",
            data=dict(switch_id=int(setting), is_public=on, game_id=game_id),
            custom_route=routes.CARD_WAPI_URL,
        )

    @deprecation.deprecated("update_settings")
    async def set_visibility(self, public: bool, *, game: typing.Optional[types.Game] = None) -> None:
        """Set your data to public or private."""
        await self.update_settings(1, public, game=game)



import enum
import typing

import pydantic

from genshin.models.model import Aliased, APIModel, Unique, prevent_enum_error

__all__ = (
    "AgentSkill",
    "AgentSkillItem",
    "DiscSetEffect",
    "WEngine",
    "ZZZAgentProperty",
    "ZZZAgentRank",
    "ZZZAwakenedPotential",
    "ZZZAwakenedPotentialInfo",
    "ZZZAwakenedPotentialSkill",
    "ZZZAwakenedPotentialSkillDetail",
    "ZZZBaseAgent",
    "ZZZDisc",
    "ZZZElementType",
    "ZZZFullAgent",
    "ZZZPartialAgent",
    "ZZZProperty",
    "ZZZPropertyType",
    "ZZZSkillType",
    "ZZZSpecialty",
)


class ZZZElementType(enum.IntEnum):
    """ZZZ element type."""

    PHYSICAL = 200
    FIRE = 201
    ICE = 202
    ELECTRIC = 203
    ETHER = 205


class ZZZSpecialty(enum.IntEnum):
    """ZZZ agent compatible specialty."""

    ATTACK = 1
    STUN = 2
    ANOMALY = 3
    SUPPORT = 4
    DEFENSE = 5
    RUPTURE = 6


class ZZZBaseAgent(APIModel, Unique):
    """ZZZ base agent model."""

    id: int  # 4 digit number
    element: ZZZElementType = Aliased("element_type")
    rarity: typing.Literal["S", "A"]
    name: str = Aliased("name_mi18n")
    full_name: str = Aliased("full_name_mi18n")
    specialty: ZZZSpecialty = Aliased("avatar_profession")
    faction_icon: str = Aliased("group_icon_path")
    rectangle_icon: str = Aliased("hollow_icon_path")
    square_icon: str = Aliased("role_square_url")


class ZZZPartialAgent(ZZZBaseAgent):
    """Character without any equipment."""

    level: int
    rank: int
    """Also known as Mindscape Cinema in-game."""

    @property
    def banner_icon(self) -> str:
        return f"https://act-webstatic.hoyoverse.com/game_record/zzzv2/role_vertical_painting/role_vertical_painting_{self.id}.png"


class ZZZPropertyType(enum.IntEnum):
    """ZZZ property type."""

    # Agent prop
    AGENT_HP = 1
    AGENT_ATK = 2
    AGENT_DEF = 3
    AGENT_IMPACT = 4
    AGENT_CRIT_RATE = 5
    AGENT_CRIT_DMG = 6
    AGENT_ANOMALY_MASTERY = 7
    AGENT_ANOMALY_PROFICIENCY = 8
    AGENT_PEN_RATIO = 9
    AGENT_ENERGY_GEN = 11
    AGENT_PEN = 232
    AGENT_SHEER_FORCE = 19
    AGENT_ADRENALINE = 20

    # Agent DMG bonus
    PHYSICAL_DMG_BONUS = 315
    FIRE_DMG_BONUS = 316
    ICE_DMG_BONUS = 317
    ELECTRIC_DMG_BONUS = 318
    ETHER_DMG_BONUS = 319

    # Disc drive and w-engine
    CRIT_RATE = 20103
    CRIT_DMG = 21103

    ANOMALY_PROFICIENCY = 31203
    ANOMALY_MASTERY = 31402
    ENERGY_REGEN = 30502
    IMPACT = 12202

    BASE_ATK = 12101
    FLAT_HP = 11103
    FLAT_ATK = 12103
    FLAT_DEF = 13103
    FLAT_PEN = 23203

    HP_PERCENT = 11102
    ATK_PERCENT = 12102
    DEF_PERCENT = 13102
    PEN_PERCENT = 23103

    DISC_PHYSICAL_DMG_BONUS = 31503
    DISC_FIRE_DMG_BONUS = 31603
    DISC_ICE_DMG_BONUS = 31703
    DISC_ELECTRIC_DMG_BONUS = 31803
    DISC_ETHER_DMG_BONUS = 31903


class ZZZProperty(APIModel):
    """A property (stat) for disc or w-engine."""

    name: str = Aliased("property_name")
    type: typing.Union[int, ZZZPropertyType] = Aliased("property_id")
    value: str = Aliased("base")

    @pydantic.field_validator("type", mode="before")
    def __cast_id(cls, v: int) -> typing.Union[int, ZZZPropertyType]:
        return prevent_enum_error(v, ZZZPropertyType)


class ZZZAgentProperty(ZZZProperty):
    """A property model, but for agents."""

    add: str
    final: str


class DiscSetEffect(APIModel):
    """A disc set effect."""

    id: int = Aliased("suit_id")
    name: str
    owned_num: int = Aliased("own")
    two_piece_description: str = Aliased("desc1")
    four_piece_description: str = Aliased("desc2")


class WEngine(APIModel):
    """A ZZZ W-engine, it's like a weapon."""

    id: int
    level: int
    name: str
    icon: str
    refinement: typing.Literal[1, 2, 3, 4, 5] = Aliased("star")
    rarity: typing.Literal["B", "A", "S"]
    properties: typing.Sequence[ZZZProperty]
    main_properties: typing.Sequence[ZZZProperty]
    effect_title: str = Aliased("talent_title")
    effect_description: str = Aliased("talent_content")
    profession: ZZZSpecialty


class ZZZDisc(APIModel):
    """A ZZZ disc, like an artifact in Genshin."""

    id: int
    level: int
    name: str
    icon: str
    rarity: typing.Literal["B", "A", "S"]
    main_properties: typing.Sequence[ZZZProperty]
    properties: typing.Sequence[ZZZProperty]
    set_effect: DiscSetEffect = Aliased("equip_suit")
    position: int = Aliased("equipment_type")


class ZZZSkillType(enum.IntEnum):
    """ZZZ agent skill type."""

    BASIC_ATTACK = 0
    DODGE = 2
    ASSIST = 6
    SPECIAL_ATTACK = 1
    CHAIN_ATTACK = 3
    CORE_SKILL = 5


class AgentSkillItem(APIModel):
    """An agent skill item."""

    title: str
    text: str


class AgentSkill(APIModel):
    """ZZZ agent skill model."""

    level: int
    type: ZZZSkillType = Aliased("skill_type")
    items: typing.Sequence[AgentSkillItem]
    """One skill can have different forms (?), so there are multiple 'items'."""


class ZZZAgentRank(APIModel):
    """ZZZ agent rank model."""

    id: int
    name: str
    description: str = Aliased("desc")
    position: int = Aliased("pos")
    unlocked: bool = Aliased("is_unlocked")


class ZZZAwakenedPotentialSkillDetail(APIModel):
    """ZZZ awakened potential skill detail model."""

    name: str = Aliased("title")
    description: str = Aliased("text")


class ZZZAwakenedPotentialSkill(APIModel):
    """ZZZ awakened potential skill model."""

    type: ZZZSkillType = Aliased("skill_type")
    title: str = Aliased("awaken_simple_info")
    skill_details: typing.Sequence[ZZZAwakenedPotentialSkillDetail] = Aliased("skill_items")


class ZZZAwakenedPotential(APIModel):
    """ZZZ awakened potential model."""

    level: int = Aliased("awaken_level")
    name: str = Aliased("level_show_name")
    buffed_skills: typing.Sequence[ZZZAwakenedPotentialSkill] = Aliased("awaken_skill_items")


class ZZZAwakenedPotentialInfo(APIModel):
    """ZZZ agent awakened potential info model."""

    has_potential: bool = Aliased("has_awaken_system")
    current: int = Aliased("awaken_level")
    max: int = Aliased("awaken_max_level")
    items: typing.Sequence[ZZZAwakenedPotential] = Aliased("skill_awaken_items")


class ZZZFullAgent(ZZZBaseAgent):
    """Character with equipment."""

    level: int
    rank: int
    """Also known as Mindscape Cinema in-game."""
    faction_name: str = Aliased("camp_name_mi18n")
    properties: typing.Sequence[ZZZAgentProperty]
    discs: typing.Sequence[ZZZDisc] = Aliased("equip")
    w_engine: typing.Optional[WEngine] = Aliased("weapon", default=None)
    skills: typing.Sequence[AgentSkill]
    ranks: typing.Sequence[ZZZAgentRank]
    """Also known as Mindscape Cinemas in-game."""
    potential: ZZZAwakenedPotentialInfo = Aliased("skill_awaken")

    banner_icon: str = Aliased("role_vertical_painting_url")

    @staticmethod
    def _extract_skin_id(url: str) -> typing.Optional[int]:
        filename = url.split("/")[-1].split(".")[0]
        parts = filename.split("_")

        if len(parts) >= 5:
            return int(parts[-1])
        return None

    @property
    def outfit_id(self) -> typing.Optional[int]:
        return self._extract_skin_id(self.banner_icon)



"""Base paginators."""

from __future__ import annotations

import abc
import asyncio
import heapq
import random
import typing

__all__ = ["BufferedPaginator", "MergedPaginator", "Paginator"]

T = typing.TypeVar("T")


async def flatten(iterable: typing.AsyncIterable[T]) -> typing.Sequence[T]:
    """Flatten an async iterable."""
    if isinstance(iterable, Paginator):
        return await iterable.flatten()  # type: ignore

    return [x async for x in iterable]


async def aiterate(iterable: typing.Iterable[T]) -> typing.AsyncIterator[T]:
    """Turn a plain iterable into an async iterator."""
    for i in iterable:
        yield i


class Paginator(typing.Generic[T], abc.ABC):
    """Base paginator."""

    __slots__ = ()

    @property
    def _repr_attributes(self) -> typing.Sequence[str]:
        """Attributes to be used in repr."""
        return [
            attribute
            for subclass in self.__class__.__mro__
            for attribute in getattr(subclass, "__slots__", ())
            if not attribute.startswith("_")
        ]

    def __repr__(self) -> str:
        kwargs = ", ".join(f"{name}={getattr(self, name, 'undefined')!r}" for name in self._repr_attributes)
        return f"{self.__class__.__name__}({kwargs})"

    def __pretty__(
        self,
        fmt: typing.Callable[[typing.Any], str],
        **kwargs: typing.Any,
    ) -> typing.Iterator[typing.Any]:
        """Devtools pretty formatting."""
        yield self.__class__.__name__
        yield "("
        yield 1

        for name in self._repr_attributes:
            yield name
            yield "="
            if hasattr(self, name):
                yield fmt(getattr(self, name))
            else:
                yield "<undefined>"

            yield 0

        yield -1
        yield ")"

    async def next(self) -> T:
        """Return the next element."""
        try:
            return await self.__anext__()
        except StopAsyncIteration:
            raise LookupError("No elements were found") from None

    def _complete(self) -> typing.NoReturn:
        """Mark paginator as complete and clear memory."""
        raise StopAsyncIteration("No more items exist in this paginator. It has been exhausted.") from None

    def __aiter__(self) -> Paginator[T]:
        return self

    async def flatten(self) -> typing.Sequence[T]:
        """Flatten the paginator."""
        return [item async for item in self]

    def __await__(self) -> typing.Generator[None, None, typing.Sequence[T]]:
        return self.flatten().__await__()

    @abc.abstractmethod
    async def __anext__(self) -> T: ...


class BasicPaginator(typing.Generic[T], Paginator[T], abc.ABC):
    """Paginator that simply iterates over an iterable."""

    __slots__ = ("iterator",)

    iterator: typing.AsyncIterator[T]
    """Underlying iterator."""

    def __init__(self, iterable: typing.Union[typing.Iterable[T], typing.AsyncIterable[T]]) -> None:
        if isinstance(iterable, typing.AsyncIterable):
            self.iterator = iterable.__aiter__()
        else:
            self.iterator = aiterate(iterable)

    async def __anext__(self) -> T:
        try:
            return await self.iterator.__anext__()
        except StopAsyncIteration:
            self._complete()


class BufferedPaginator(typing.Generic[T], Paginator[T], abc.ABC):
    """Paginator with a support for buffers."""

    __slots__ = ("limit", "_buffer", "_counter")

    limit: typing.Optional[int]
    """Limit of items to be yielded."""

    _buffer: typing.Optional[typing.Iterator[T]]
    """Item buffer. If none then exhausted."""

    _counter: int
    """Amount of yielded items so far. No guarantee to be synchronized."""

    def __init__(self, *, limit: typing.Optional[int] = None) -> None:
        self.limit = limit

        self._buffer = iter(())
        self._counter = 0

    @property
    def exhausted(self) -> bool:
        """Whether all pages have been fetched."""
        return self._buffer is None

    def _complete(self) -> typing.NoReturn:
        self._buffer = None

        super()._complete()
        raise  # pyright bug

    @abc.abstractmethod
    async def next_page(self) -> typing.Optional[typing.Iterable[T]]:
        """Get the next page of the paginator."""

    async def __anext__(self) -> T:
        if not self._buffer:
            self._complete()

        if self.limit and self._counter >= self.limit:
            self._complete()

        self._counter += 1

        try:
            return next(self._buffer)
        except StopIteration:
            pass

        buffer = await self.next_page()
        if not buffer:
            self._complete()

        self._buffer = iter(buffer)
        return next(self._buffer)


class MergedPaginator(typing.Generic[T], Paginator[T]):
    """A paginator merging a collection of iterators."""

    __slots__ = ("iterators", "_heap", "limit", "_key", "_prepared", "_counter")

    # TODO: Use named tuples for the heap

    iterators: typing.Sequence[typing.AsyncIterator[T]]
    """Entry iterators.

    Only used as pointers to a heap.
    """

    _heap: list[tuple[typing.Any, int, T, typing.AsyncIterator[T]]]
    """Underlying heap queue.

    List of (comparable, unique order id, value, iterator)
    """

    limit: typing.Optional[int]
    """Limit of items to be yielded"""

    _key: typing.Optional[typing.Callable[[T], typing.Any]]
    """Sorting key."""

    _prepared: bool
    """Whether the paginator is prepared"""

    _counter: int
    """Amount of yielded items so far. No guarantee to be synchronized."""

    def __init__(
        self,
        iterables: typing.Collection[typing.AsyncIterable[T]],
        *,
        key: typing.Optional[typing.Callable[[T], typing.Any]] = None,
        limit: typing.Optional[int] = None,
    ) -> None:
        self.iterators = [iterable.__aiter__() for iterable in iterables]
        self._key = key
        self.limit = limit

        self._prepared = False
        self._counter = 0

    def _complete(self) -> typing.NoReturn:
        """Mark paginator as complete and clear memory."""
        # free memory in heaps
        self._heap = []
        self.iterators = []

        super()._complete()
        raise  # pyright bug

    def _create_heap_item(
        self,
        value: T,
        iterator: typing.AsyncIterator[T],
        order: typing.Optional[int] = None,
    ) -> tuple[typing.Any, int, T, typing.AsyncIterator[T]]:
        """Create a new item for the heap queue."""
        sort_value = self._key(value) if self._key else value
        if order is None:
            order = random.getrandbits(16)

        return (sort_value, order, value, iterator)

    async def _prepare(self) -> None:
        """Prepare the heap queue by filling it with initial values."""
        coros = (it.__anext__() for it in self.iterators)
        first_values = await asyncio.gather(*coros, return_exceptions=True)

        self._heap = []
        for order, (it, value) in enumerate(zip(self.iterators, first_values)):
            if isinstance(value, BaseException):
                if isinstance(value, StopAsyncIteration):
                    continue

                raise value

            heapq.heappush(self._heap, self._create_heap_item(value, iterator=it, order=order))

        self._prepared = True

    async def __anext__(self) -> T:
        if not self._prepared:
            await self._prepare()

        if not self._heap:
            self._complete()

        if self.limit and self._counter >= self.limit:
            self._complete()

        self._counter += 1

        _, order, value, it = self._heap[0]

        try:
            new_value = await it.__anext__()
        except StopAsyncIteration:
            heapq.heappop(self._heap)
            return value

        heapq.heapreplace(self._heap, self._create_heap_item(new_value, iterator=it, order=order))

        return value

    async def flatten(self, *, lazy: bool = False) -> typing.Sequence[T]:
        """Flatten the paginator."""
        if self.limit is not None and lazy:
            return [item async for item in self]

        coros = (flatten(i) for i in self.iterators)
        lists: typing.Sequence[typing.Sequence[T]] = await asyncio.gather(*coros)  # pyright: ignore

        return list(heapq.merge(*lists, key=self._key))[: self.limit]  # pyright: ignore



import typing

import pydantic

from genshin.models.model import Aliased, APIModel
from genshin.models.zzz.chronicle.events import ZZZGachaEventType

__all__ = (
    "ZZZGachaBannerInfo",
    "ZZZGachaCurrencies",
    "ZZZGachaInfo",
)


class ZZZGachaCurrencies(APIModel):
    """ZZZ gacha currency counts."""

    monochrome: int
    polychrome: int
    encrypted_master_tape: int
    master_tape: int
    boopon: int


class ZZZGachaBannerInfo(APIModel):
    """ZZZ gacha banner info model."""

    type: ZZZGachaEventType = Aliased("gacha_type")
    pity: int = Aliased("more_s_need_cnt")
    """Number of pulls needed until guaranteed S."""


class ZZZGachaInfo(APIModel):
    """ZZZ gacha info model."""

    currencies: ZZZGachaCurrencies
    banners: list[ZZZGachaBannerInfo] = Aliased("gacha_info_list")

    @pydantic.model_validator(mode="before")
    @classmethod
    def __convert_currencies(cls, data: dict[str, typing.Any]) -> dict[str, typing.Any]:
        currency_mapping = {
            "GACHA_TICKET_TYPE_RECHARGE_MONOCHROME": "monochrome",
            "GACHA_TICKET_TYPE_POLYCHROME": "polychrome",
            "GACHA_TICKET_TYPE_ENCRYPTED_MASTER_TAPE": "encrypted_master_tape",
            "GACHA_TICKET_TYPE_MASTER_TAPE": "master_tape",
            "GACHA_TICKET_TYPE_BOOPON": "boopon",
        }

        tickets = data.get("tickets", [])
        converted_currencies = {
            currency_mapping[ticket["ticket_type"]]: ticket["ticket_cnt"]
            for ticket in tickets
            if ticket["ticket_type"] in currency_mapping
        }
        data["currencies"] = converted_currencies
        return data



"""HSR Lineup Simulator."""

import datetime
import typing
from enum import Enum

import pydantic

from genshin.models.model import Aliased, APIModel, DateTime

from . import character

__all__ = (
    "APCShadowLineup",
    "APCShadowLineupResponse",
    "APCShadowSchedule",
    "LineupDetail",
    "MOCSchedule",
    "PureFictionLineup",
    "PureFictionLineupResponse",
    "PureFictionSchedule",
    "StarRailGameMode",
    "StarRailGameModeBuff",
    "StarRailGameModeFloor",
    "StarRailGameModeSchedule",
    "StarRailGameModeType",
    "StarRailLineup",
    "StarRailLineupPlayer",
    "StarRailLineupResponse",
)


class StarRailGameModeType(str, Enum):
    """HSR lineup game mode enum."""

    MOC = "Chasm"
    """Memory of Chaos."""
    PURE_FICTION = "Story"
    """Pure Fiction."""
    APC_SHADOW = "Boss"
    """Apocalyptic Shadow."""
    ANOMALY_ARBITRATION = "Peak"
    """Anomaly Arbitration."""


class StarRailGameModeFloor(APIModel):
    """A floor of a game mode, like MOC Stage 1."""

    id: int
    name: str
    floor: int

    @pydantic.model_validator(mode="before")
    @classmethod
    def __extract_floor(cls, v: dict[str, typing.Any]) -> dict[str, typing.Any]:
        extend = v["extend"]
        v["floor"] = extend["floor"]
        return v


class StarRailGameMode(APIModel):
    """HSR game mode, like MOC."""

    id: int
    name: str
    floors: typing.Sequence[StarRailGameModeFloor]
    type: StarRailGameModeType = Aliased("root_type")

    @pydantic.model_validator(mode="before")
    @classmethod
    def __unnest_children(cls, v: dict[str, typing.Any]) -> dict[str, typing.Any]:
        """Unnest the children field."""
        if not (children := v.get("children")):
            msg = "Missing 'children' field in HSRGameMode."
            raise ValueError(msg)

        v["floors"] = children[0]["children"]
        v["name"] = children[0]["name"]
        return v


class StarRailGameModeBuff(APIModel):
    """HSR game mode buff."""

    name: str
    desc: str
    icon: typing.Optional[str] = None
    simple_desc: typing.Optional[str] = None


class StarRailGameModeSchedule(APIModel):
    """Base HSR lineup game mode schedule."""

    id: int = Aliased("group_id")
    name: str = Aliased("name_mi18n")
    start_time: datetime.datetime = Aliased("schedule_begin_time")
    end_time: datetime.datetime = Aliased("schedule_end_time")

    is_now: bool
    is_show: bool
    floor_nums: int


class APCShadowSchedule(StarRailGameModeSchedule):
    """Apocalyptic Shadow schedule."""

    buff: StarRailGameModeBuff
    node1_buffs: typing.Sequence[StarRailGameModeBuff] = Aliased("addition_buff_1")
    node2_buffs: typing.Sequence[StarRailGameModeBuff] = Aliased("addition_buff_2")

    @pydantic.model_validator(mode="before")
    @classmethod
    def __convert_buffs(cls, v: dict[str, typing.Any]) -> dict[str, typing.Any]:
        v["buff"] = {"name": v["maze_buff_name_mi18n"], "desc": v["maze_buff_desc_mi18n"], "icon": v["buff_icon"]}
        return v


class MOCSchedule(StarRailGameModeSchedule):
    """Memory of Chaos schedule."""

    buff: StarRailGameModeBuff

    @pydantic.model_validator(mode="before")
    @classmethod
    def __convert_buffs(cls, v: dict[str, typing.Any]) -> dict[str, typing.Any]:
        v["buff"] = {"name": v["maze_buff_name_mi18n"], "desc": v["maze_buff_desc_mi18n"]}
        return v


class PureFictionSchedule(StarRailGameModeSchedule):
    """Pure Future schedule."""

    buffs: typing.Sequence[StarRailGameModeBuff] = Aliased("addition_buff")
    default_buffs: typing.Sequence[StarRailGameModeBuff] = Aliased("sub_maze_buff_list")


class StarRailLineupPlayer(APIModel):
    """A HSR lineup player."""

    uid: str
    nickname: str
    avatar_url: str
    level: int


class StarRailLineup(APIModel):
    """A HSR lineup."""

    id: str
    player: StarRailLineupPlayer
    type: StarRailGameModeType = Aliased("lineup_type")

    title: str
    description: str
    likes: int = Aliased("like_cnt")
    comments: int = Aliased("comment_cnt")
    favorites: int = Aliased("favour_cnt")
    views: int = Aliased("view_cnt")

    created_at: datetime.datetime
    last_edited_at: datetime.datetime = Aliased("last_edit")

    stars: int = Aliased("star_num")
    cycles_taken: int = Aliased("round_num")

    characters: typing.Sequence[typing.Sequence[character.StarRailLineupCharacter]] = Aliased("avatar_group")

    @pydantic.field_validator("characters", mode="before")
    @classmethod
    def __unnest_characters(
        cls, v: typing.Sequence[dict[str, typing.Any]]
    ) -> typing.Sequence[typing.Sequence[dict[str, typing.Any]]]:
        # Merge relic info from "group"
        for g in v:
            group = g["group"]

            for avatar in g["avatar_details"]:
                group_avatar = next((a for a in group if a["item_id"] == str(avatar["id"])), None)
                if not group_avatar:
                    continue

                relics = group_avatar.get("relics", []) + group_avatar.get("relic_sides", [])

                for relic in avatar.get("relics", []):
                    group_avatar_relic = next((r for r in relics if r["item_id"] == str(relic["id"])), None)
                    if not group_avatar_relic:
                        continue

                    relic.update(group_avatar_relic)

        return [g["avatar_details"] for g in v]

    @pydantic.model_validator(mode="before")
    @classmethod
    def __nest_player(cls, v: dict[str, typing.Any]) -> dict[str, typing.Any]:
        """Nest the player field."""
        v["player"] = {
            "uid": v["account_uid"],
            "nickname": v["nickname"],
            "avatar_url": v["avatar_url"],
            "level": v["game_level"],
        }
        return v


class LineupDetail(APIModel):
    """Detail for a lineup."""

    buffs: typing.Sequence[StarRailGameModeBuff] = Aliased("buff_list")
    points: int


class PureFictionLineup(StarRailLineup):
    """A Pure Fiction lineup."""

    detail: LineupDetail = Aliased("story_info")
    node1_challenged_at: DateTime = Aliased("challenge_time")
    node2_challenged_at: DateTime = Aliased("challenge_time_node2")


class APCShadowLineup(StarRailLineup):
    """A Apocalyptic Shadow lineup."""

    detail: LineupDetail = Aliased("boss_info")
    node1_challenged_at: DateTime = Aliased("challenge_time")
    node2_challenged_at: DateTime = Aliased("challenge_time_node2")


class StarRailLineupResponse(APIModel):
    """Response for HSR lineups."""

    lineups: typing.Sequence[StarRailLineup] = Aliased("list")
    next_page_token: str


class PureFictionLineupResponse(APIModel):
    """Response for Pure Fiction lineups."""

    lineups: typing.Sequence[PureFictionLineup] = Aliased("list")
    next_page_token: str


class APCShadowLineupResponse(APIModel):
    """Response for Apocalyptic Shadow lineups."""

    lineups: typing.Sequence[APCShadowLineup] = Aliased("list")
    next_page_token: str



"""Miyoushe QR Code Models"""

import enum

import pydantic

__all__ = ["QRCodeCreationResult", "QRCodeStatus"]


class QRCodeStatus(enum.Enum):
    """QR code check status."""

    CREATED = "Created"
    SCANNED = "Scanned"
    CONFIRMED = "Confirmed"


class QRCodeCreationResult(pydantic.BaseModel):
    """QR code creation result."""

    ticket: str
    url: str



"""ZZZ data overview models."""

import typing

import pydantic

from genshin.models.model import Aliased, APIModel

from ..character import ZZZPartialAgent

__all__ = ("HIACoin", "ZZZBaseBangboo", "ZZZCatNote", "ZZZGameData", "ZZZMedal", "ZZZStats", "ZZZUserStats")


class HIACoin(APIModel):
    """HIACoin model."""

    num: int
    name: str
    icon: str = Aliased("url")


class ZZZStats(APIModel):
    """ZZZ data overview stats."""

    active_days: int
    character_num: int = Aliased("avatar_num")
    inter_knot_reputation: str = Aliased("world_level_name")
    shiyu_defense_frontiers: int = Aliased("cur_period_zone_layer_count")
    bangboo_obtained: int = Aliased("buddy_num")
    achievement_count: int
    hia_coin: typing.Optional[HIACoin] = Aliased("commemorative_coins_list")

    @pydantic.field_validator("hia_coin", mode="before")
    def __unnest_hia_coin(
        cls, v: typing.List[typing.Dict[str, typing.Any]]
    ) -> typing.Optional[typing.Dict[str, typing.Any]]:
        return v[0] if v else None


class ZZZCatNote(APIModel):
    """ZZZ Cat note model."""

    icon: str
    id: int
    is_lock: bool
    name: str
    num: int
    total: int


class ZZZMedal(APIModel):
    """ZZZ player medal model."""

    icon: str = Aliased("medal_icon")
    number: int
    id: int = Aliased("medal_id")
    type: str = Aliased("medal_type")
    name: str
    is_show: bool


class ZZZGameData(APIModel):
    """ZZZ game data model."""

    personal_title: str
    title_main_color: str
    title_bottom_color: str
    title_bg_url: str
    medal_icons: typing.Sequence[str] = Aliased("medal_list")
    card_url: str
    medals: typing.Sequence[ZZZMedal] = Aliased("all_medal_list")


class ZZZBaseBangboo(APIModel):
    """Base bangboo (buddy) model."""

    id: int
    name: str
    rarity: typing.Literal["S", "A"]
    level: int
    star: int
    icon: str = pydantic.Field(
        validation_alias=pydantic.aliases.AliasChoices("bangboo_square_url", "bangboo_rectangle_url")
    )


class ZZZUserStats(APIModel):
    """Zenless Zone Zero user model."""

    stats: ZZZStats
    agents: typing.Sequence[ZZZPartialAgent] = Aliased("avatar_list")
    bangboos: typing.Sequence[ZZZBaseBangboo] = Aliased("buddy_list")
    in_game_avatar: str = Aliased("cur_head_icon_url")
    cat_notes: typing.Sequence[ZZZCatNote] = Aliased("cat_notes_list")
    in_game_data: ZZZGameData = Aliased("game_data_show")



import datetime
import enum
import typing

import pydantic

from genshin.models.genshin import character
from genshin.models.model import Aliased, APIModel, TZDateTime

__all__ = (
    "Act",
    "ActCharacter",
    "BattleStatCharacter",
    "ImgTheater",
    "ImgTheaterData",
    "TheaterBattleStats",
    "TheaterBuff",
    "TheaterCharaType",
    "TheaterDifficulty",
    "TheaterSchedule",
    "TheaterStats",
)


class TheaterCharaType(enum.IntEnum):
    """The type of character in the context of the imaginarium theater gamemode."""

    NORMAL = 1
    TRIAL = 2
    SUPPORT = 3


class TheaterDifficulty(enum.IntEnum):
    """The difficulty of the imaginarium theater data."""

    UNKNOWN = 0
    EASY = 1
    NORMAL = 2
    HARD = 3
    VISIONARY = 4
    ARCANA_CHALLENGE = 5


class ActCharacter(character.BaseCharacter):
    """A character in an act."""

    type: TheaterCharaType = Aliased("avatar_type")
    level: int


class TheaterBuff(APIModel):
    """Represents either a 'mystery cache' or a 'wondrous boom'."""

    icon: str
    name: str
    description: str = Aliased("desc")
    received_audience_support: bool = Aliased("is_enhanced")
    """Whether external audience support is received."""
    id: int


class Act(APIModel):
    """One act in the theater."""

    characters: typing.Sequence[ActCharacter] = Aliased("avatars")
    mystery_caches: typing.Sequence[TheaterBuff] = Aliased("choice_cards")
    wondroud_booms: typing.Sequence[TheaterBuff] = Aliased("buffs")
    medal_obtained: bool = Aliased("is_get_medal")
    round_id: int
    finish_time: int  # As timestamp
    finish_datetime: TZDateTime = Aliased("finish_date_time")
    is_arcana: bool = Aliased("is_tarot", default=False)
    arcana_number: typing.Optional[int] = Aliased("tarot_serial_no", default=None)

    @pydantic.field_validator("finish_datetime", mode="before")
    def __parse_datetime(cls, value: typing.Mapping[str, typing.Any]) -> datetime.datetime:
        return datetime.datetime(**value)


class TheaterStats(APIModel):
    """Imaginarium theater stats."""

    difficulty: TheaterDifficulty = Aliased("difficulty_id")
    best_record: int = Aliased("max_round_id")
    """The maximum act the player has reached."""
    heraldry: int  # Not sure what this is
    star_challenge_stellas: typing.Sequence[bool] = Aliased("get_medal_round_list")
    """Whether the player has obtained the medal for each act."""
    fantasia_flowers_used: int = Aliased("coin_num")
    """The number of Fantasia Flowers used."""
    audience_support_trigger_num: int = Aliased("avatar_bonus_num")
    """The number of external audience support triggers."""
    player_assists: int = Aliased("rent_cnt")
    """The number of supporting cast characters assisting other players."""
    medal_num: int
    """The number of medals the player has obtained."""


class TheaterSchedule(APIModel):
    """Imaginarium theater schedule."""

    start_time: int  # As timestamp
    end_time: int  # As timestamp
    schedule_type: int  # Not sure what this is
    id: int = Aliased("schedule_id")
    start_datetime: TZDateTime = Aliased("start_date_time")
    end_datetime: TZDateTime = Aliased("end_date_time")

    @pydantic.field_validator("start_datetime", "end_datetime", mode="before")
    def __parse_datetime(cls, value: typing.Mapping[str, typing.Any]) -> datetime.datetime:
        return datetime.datetime(**value)


class BattleStatCharacter(APIModel):
    """Imaginarium theater battle statistic character."""

    id: int = Aliased("avatar_id")
    icon: str = Aliased("avatar_icon")
    value: int
    rarity: int

    @pydantic.field_validator("value", mode="before")
    def __intify_value(cls, value: str) -> int:
        if not value:
            return 0
        return int(value)


class TheaterBattleStats(APIModel):
    """Imaginarium theater battle statistics."""

    max_defeat_character: typing.Optional[BattleStatCharacter] = Aliased("max_defeat_avatar", default=None)
    max_damage_character: typing.Optional[BattleStatCharacter] = Aliased("max_damage_avatar", default=None)
    max_take_damage_character: typing.Optional[BattleStatCharacter] = Aliased("max_take_damage_avatar", default=None)
    fastest_character_list: typing.Sequence[BattleStatCharacter] = Aliased("shortest_avatar_list")
    total_cast_seconds: int = Aliased("total_use_time")

    @pydantic.field_validator(
        "max_defeat_character", "max_damage_character", "max_take_damage_character", mode="before"
    )
    def __none_if_empty(cls, value: dict[str, typing.Any]) -> typing.Any:
        if not value or value.get("avatar_id", 0) == 0:
            return None
        return value


class ImgTheaterData(APIModel):
    """Imaginarium theater data."""

    acts: typing.Sequence[Act] = Aliased(alias="rounds_data")
    backup_characters: typing.Sequence[ActCharacter] = Aliased("backup_avatars")  # Not sure what this is
    stats: TheaterStats = Aliased(alias="stat")
    schedule: TheaterSchedule
    has_data: bool
    has_detail_data: bool
    battle_stats: typing.Optional[TheaterBattleStats] = Aliased("fight_statisic", default=None)

    @pydantic.model_validator(mode="before")
    def __unnest_detail(cls, values: dict[str, typing.Any]) -> dict[str, typing.Any]:
        detail: typing.Optional[dict[str, typing.Any]] = values.get("detail")
        values["rounds_data"] = detail.get("rounds_data", []) if detail is not None else []
        values["backup_avatars"] = detail.get("backup_avatars", []) if detail is not None else []
        values["fight_statisic"] = detail.get("fight_statisic", None) if detail is not None else None
        return values


class ImgTheater(APIModel):
    """Imaginarium theater."""

    datas: typing.Sequence[ImgTheaterData] = Aliased("data")
    unlocked: bool = Aliased("is_unlock")



"""Stygian Onslaught models."""

import enum
import typing

import pydantic

from genshin.models.model import Aliased, APIModel, DateTime, LevelField

__all__ = (
    "HardChallenge",
    "HardChallengeBestCharacter",
    "HardChallengeBestCharacterType",
    "HardChallengeBestRecord",
    "HardChallengeChallenge",
    "HardChallengeCharacter",
    "HardChallengeData",
    "HardChallengeEnemy",
    "HardChallengeEnemyTag",
    "HardChallengeSeason",
    "HardChallengeTagElement",
    "HardChallengeTagType",
)


class HardChallengeBestCharacterType(enum.IntEnum):
    """Type of best character in Stygian Onslaught."""

    STRIKE = 1
    """Strongest single strike."""
    DAMAGE = 2
    """Highest total damage dealt."""


class HardChallengeTagType(enum.IntEnum):
    """Type of tag in Stygian Onslaught."""

    DISADVANTAGE = 0
    ADVANTAGE = 1


class HardChallengeTagElement(enum.Enum):
    """Element of tag in Stygian Onslaught."""

    CRYO = "{SPRITE_PRESET#11001}"
    HYDRO = "{SPRITE_PRESET#11002}"
    PYRO = "{SPRITE_PRESET#11003}"
    DENDRO = "{SPRITE_PRESET#11007}"


class HardChallengeSeason(APIModel):
    """A season of Stygian Onslaught."""

    id: int = Aliased("schedule_id")
    name: str
    start_at: DateTime = Aliased("start_date_time")
    end_at: DateTime = Aliased("end_date_time")


class HardChallengeBestRecord(APIModel):
    """Best record for a Stygian Onslaught season."""

    difficulty: int
    time_used: int = Aliased("second")
    """Time used for challenge in seconds."""
    icon: str
    """Badge icon filename."""

    @pydantic.field_validator("icon")
    @classmethod
    def __parse_icon(cls, v: str) -> str:
        return v.split(",")[-1]


class HardChallengeCharacter(APIModel):
    """A character in Stygian Onslaught."""

    id: int = Aliased("avatar_id")
    name: str
    element: str
    icon: str = Aliased("image")
    level: LevelField
    rarity: int
    constellation: int = Aliased("rank")


class HardChallengeBestCharacter(APIModel):
    """Best character in Stygian Onslaught."""

    id: int = Aliased("avatar_id")
    side_icon: str
    value: str = Aliased("dps")
    type: HardChallengeBestCharacterType


class HardChallengeEnemyTag(APIModel):
    """An enemy tag in Stygian Onslaught."""

    type: HardChallengeTagType
    elements: list[HardChallengeTagElement]

    @pydantic.model_validator(mode="before")
    @classmethod
    def __parse_elements(cls, v: dict[str, typing.Any]) -> dict[str, typing.Any]:
        """Parse elements from string to HardChallengeTagElement enum."""
        desc = v.get("desc", "")
        for element in HardChallengeTagElement:
            if element.value in desc:
                v.setdefault("elements", []).append(element)
        return v


class HardChallengeEnemy(APIModel):
    """An enemy in Stygian Onslaught."""

    id: int = Aliased("monster_id")
    name: str
    level: LevelField
    icon: str
    descriptions: list[str] = Aliased("desc")
    tags: list[HardChallengeEnemyTag]


class HardChallengeChallenge(APIModel):
    """A single/multi player challenge in Stygian Onslaught."""

    name: str
    time_used: int = Aliased("second")
    """Time used for challenge in seconds."""
    team: list[HardChallengeCharacter] = Aliased("teams")
    best_characters: list[HardChallengeBestCharacter] = Aliased("best_avatar")
    enemy: HardChallengeEnemy = Aliased("monster")


class HardChallengeData(APIModel):
    """Stygian Onslaught single/multi player data."""

    best_record: typing.Optional[HardChallengeBestRecord] = Aliased("best")
    challenges: list[HardChallengeChallenge] = Aliased("challenge")
    has_data: bool


class HardChallenge(APIModel):
    """Stygian Onslaught data."""

    season: HardChallengeSeason = Aliased("schedule")
    single_player: HardChallengeData = Aliased("single")
    multi_player: HardChallengeData = Aliased("mp")



import typing
from enum import Enum

import pydantic

from genshin.models.model import Aliased, APIModel, UnixDateTime, prevent_enum_error
from genshin.models.zzz.character import ZZZElementType, ZZZSpecialty

__all__ = (
    "ZZZCharacterGachaEvent",
    "ZZZEvent",
    "ZZZEventStatus",
    "ZZZGachaCalendar",
    "ZZZGachaEvent",
    "ZZZGachaEventCharacter",
    "ZZZGachaEventStatus",
    "ZZZGachaEventType",
    "ZZZGachaEventWeapon",
    "ZZZWeaponGachaEvent",
)


class ZZZEventStatus(Enum):
    """ZZZ event status enum."""

    IN_PROGRESS = "STATE_IN_PROGRESS"
    COMPLETED = "STATE_COMPLETED"
    NOT_STARTED = "STATE_NOT_START"


class ZZZGachaEventStatus(Enum):
    """ZZZ gacha event status enum."""

    IN_PROGRESS = "GACHA_STATE_IN_PROGRESS"
    COMPLETED = "GACHA_STATE_COMPLETED"
    NOT_STARTED = "GACHA_STATE_NOT_START"


class ZZZGachaEventType(Enum):
    """ZZZ gacha event type enum."""

    CHARACTER = "GACHA_TYPE_CHARACTER_UP"
    WEAPON = "GACHA_TYPE_WEAPON_UP"
    PERMANENT = "GACHA_TYPE_PERMANENT"
    BANGBOO = "GACHA_TYPE_BANGBOO"
    EXCLUSIVE_RESCREENING = "GACHA_TYPE_CHARACTER_RETURN"
    REVERBERATION = "GACHA_TYPE_WEAPON_RETURN"


class ZZZEvent(APIModel):
    """ZZZ event calendar event."""

    id: int = Aliased("activity_id")
    status: typing.Union[ZZZEventStatus, str] = Aliased("state")
    name: str

    obtained_monochromes: int = Aliased("monochrome_got_cnt")
    max_monochromes: int = Aliased("monochrome_cnt")

    start: UnixDateTime = Aliased("start_ts")
    end: UnixDateTime = Aliased("end_ts")

    seconds_until_start: int = Aliased("left_start_ts")
    seconds_until_end: int = Aliased("left_end_ts")

    @pydantic.field_validator("status", mode="before")
    def __validate_status(cls, v: str) -> typing.Union[ZZZEventStatus, str]:
        return prevent_enum_error(v, ZZZEventStatus)


class ZZZGachaEvent(APIModel):
    """ZZZ gacha event model."""

    type: typing.Union[ZZZGachaEventType, str] = Aliased("gacha_type")
    status: typing.Union[ZZZGachaEventStatus, str] = Aliased("gacha_state")
    version: str
    """Like 2.4, 2.5."""

    start: UnixDateTime = Aliased("start_ts")
    end: UnixDateTime = Aliased("end_ts")

    seconds_until_start: int = Aliased("left_start_ts")
    seconds_until_end: int = Aliased("left_end_ts")

    # No clue what these are
    sup_lock_show: bool
    insurance_id: int

    @pydantic.field_validator("type", mode="before")
    def __validate_type(cls, v: str) -> typing.Union[ZZZGachaEventType, str]:
        return prevent_enum_error(v, ZZZGachaEventType)

    @pydantic.field_validator("status", mode="before")
    def __validate_status(cls, v: str) -> typing.Union[ZZZGachaEventStatus, str]:
        return prevent_enum_error(v, ZZZGachaEventStatus)


class ZZZGachaEventCharacter(APIModel):
    """ZZZ gacha event character."""

    id: int = Aliased("avatar_id")
    name: str = Aliased("avatar_name")
    full_name: str

    rarity: typing.Literal["S", "A"]
    icon: str
    specialty: ZZZSpecialty = Aliased("avatar_profession")
    element: ZZZElementType = Aliased("avatar_element_type")
    sub_element: int = Aliased("avatar_sub_element_type")  # Probably like Miyabi's Frost

    wiki_url: str
    jump_cultivate: bool
    is_forward: bool
    show_upon: bool


class ZZZGachaEventWeapon(APIModel):
    """ZZZ gacha event weapon."""

    id: int = Aliased("weapon_id")
    rarity: typing.Literal["S", "A"]
    icon: str
    specialty: ZZZSpecialty = Aliased("profession")

    skill_name: str = Aliased("talent_title")
    skill_description: str = Aliased("talent_content")

    wiki_url: str
    show_upon: bool


class ZZZCharacterGachaEvent(ZZZGachaEvent):
    """ZZZ gacha event for characters."""

    characters: typing.Sequence[ZZZGachaEventCharacter] = Aliased("avatar_list")


class ZZZWeaponGachaEvent(ZZZGachaEvent):
    """ZZZ gacha event for weapons."""

    weapons: typing.Sequence[ZZZGachaEventWeapon] = Aliased("weapon_list")


class ZZZGachaCalendar(APIModel):
    """ZZZ gacha calendar model."""

    characters: typing.Sequence[ZZZCharacterGachaEvent] = Aliased("avatar_gacha_schedule_list")
    weapons: typing.Sequence[ZZZWeaponGachaEvent] = Aliased("weapon_gacha_schedule_list")



"""Base hoyolab APIModels."""

from __future__ import annotations

import enum
import re
import typing

import pydantic

from genshin import types
from genshin.models.model import Aliased, APIModel, Unique

__all__ = [
    "FullHoyolabUser",
    "Gender",
    "GenshinAccount",
    "HoyolabUserCertification",
    "HoyolabUserLevel",
    "PartialHoyolabUser",
    "RecordCard",
    "RecordCardData",
    "RecordCardSetting",
    "RecordCardSettingType",
    "UserInfo",
]


class GenshinAccount(APIModel):
    """Genshin account."""

    game_biz: str
    uid: int = Aliased("game_uid")
    level: int
    nickname: str
    server: str = Aliased("region")
    server_name: str = Aliased("region_name")

    @property
    def game(self) -> types.Game:
        if "hk4e" in self.game_biz:
            return types.Game.GENSHIN
        if "bh3" in self.game_biz:
            return types.Game.HONKAI
        if "hkrpg" in self.game_biz:
            return types.Game.STARRAIL
        if "nap" in self.game_biz:
            return types.Game.ZZZ
        if "nxx" in self.game_biz:
            return types.Game.TOT

        try:
            return types.Game(self.game_biz)
        except ValueError:
            return self.game_biz  # type: ignore


class UserInfo(APIModel):
    """Chronicle user info."""

    nickname: str
    server: str = Aliased("region")
    level: int
    icon: str = Aliased("AvatarUrl")
    in_game_avatar: typing.Optional[str] = Aliased("game_head_icon", default=None)


class RecordCardData(APIModel):
    """Data entry of a record card."""

    name: str
    value: str


class RecordCardSetting(APIModel, Unique):
    """Privacy setting of a record card."""

    id: int = Aliased("switch_id")
    description: str = Aliased("switch_name")
    public: bool = Aliased("is_public")


class RecordCardSettingType(enum.IntEnum):
    """Privacy setting of a record card."""

    SHOW_CHRONICLE = 1
    SHOW_CHARACTER_DETAILS = 2
    ENABLE_REAL_TIME_NOTES = 3


class Gender(enum.IntEnum):
    """Gender used on hoyolab."""

    unknown = 0
    male = 1
    female = 2
    other = 3


class PartialHoyolabUser(APIModel):
    """Partial hoyolab user from a search result."""

    hoyolab_id: int = Aliased("uid")
    nickname: str
    introduction: str = Aliased("introduce")
    avatar_id: int = Aliased("avatar")
    gender: Gender
    icon: str = Aliased("avatar_url")

    @pydantic.field_validator("nickname")
    def __remove_highlight(cls, v: str) -> str:
        return re.sub(r"<.+?>", "", v)


class HoyolabUserCertification(APIModel):
    """Hoyolab user certification.

    For example artist's type is 2.
    """

    icon_url: typing.Optional[str] = None
    description: typing.Optional[str] = Aliased("desc", default=None)
    type: int


class HoyolabUserLevel(APIModel):
    """Hoyolab user level."""

    level: int
    exp: int
    level_desc: str
    bg_color: str
    bg_image: str


class FullHoyolabUser(PartialHoyolabUser):
    """Full hoyolab user.

    Not actually full, but most of the data is useless.
    """

    certification: typing.Optional[HoyolabUserCertification] = None
    level: typing.Optional[HoyolabUserLevel] = None
    pendant_url: str = Aliased("pendant")
    bg_url: typing.Optional[str] = None
    pc_bg_url: typing.Optional[str] = None


class RecordCard(GenshinAccount):
    """Hoyolab record card."""

    def __new__(cls, **kwargs: typing.Any) -> RecordCard:
        """Create the appropriate record card."""
        game_id = kwargs.get("game_id", 0)
        if game_id == 1:
            cls = HonkaiRecordCard
        elif game_id == 2:
            cls = GenshinRecordCard
        elif game_id == 6:
            cls = StarRailRecodeCard
        elif game_id == 8:
            cls = ZZZRecordCard

        return super().__new__(cls)  # type: ignore

    game_id: int
    game_biz: str = ""
    game_name: str
    game_logo: str = Aliased("logo")
    uid: int = Aliased("game_role_id")

    data: typing.Sequence[RecordCardData]
    settings: typing.Sequence[RecordCardSetting] = Aliased("data_switches")

    public: bool = Aliased("is_public")
    background_image: str
    has_uid: bool = Aliased("has_role")
    url: str

    def as_dict(self) -> dict[str, typing.Any]:
        """Return data as a dictionary."""
        return {d.name: (int(d.value) if d.value.isdigit() else d.value) for d in self.data}


class GenshinRecordCard(RecordCard):
    """Genshin record card."""

    @property
    def game(self) -> types.Game:
        return types.Game.GENSHIN

    @property
    def days_active(self) -> int:
        return int(self.data[0].value)

    @property
    def characters(self) -> int:
        return int(self.data[1].value)

    @property
    def achievements(self) -> int:
        return int(self.data[2].value)

    @property
    def spiral_abyss(self) -> str:
        return self.data[3].value


class HonkaiRecordCard(RecordCard):
    """Honkai record card."""

    @property
    def game(self) -> types.Game:
        return types.Game.HONKAI

    @property
    def days_active(self) -> int:
        return int(self.data[0].value)

    @property
    def stigmata(self) -> int:
        return int(self.data[1].value)

    @property
    def battlesuits(self) -> int:
        return int(self.data[2].value)

    @property
    def outfits(self) -> int:
        return int(self.data[3].value)


class StarRailRecodeCard(RecordCard):
    """Star rail record card."""

    @property
    def game(self) -> types.Game:
        return types.Game.STARRAIL

    @property
    def days_active(self) -> int:
        return int(self.data[0].value)

    @property
    def characters(self) -> int:
        return int(self.data[1].value)

    @property
    def achievements(self) -> int:
        return int(self.data[2].value)

    @property
    def chests(self) -> int:
        return int(self.data[3].value)


class ZZZRecordCard(RecordCard):
    """ZZZ record card."""

    @property
    def game(self) -> types.Game:
        return types.Game.ZZZ

    @property
    def days_active(self) -> int:
        return int(self.data[0].value)

    @property
    def inter_knot_reputation(self) -> str:
        return self.data[1].value

    @property
    def agents_recruited(self) -> int:
        return int(self.data[2].value)

    @property
    def bangboo_obtained(self) -> int:
        return int(self.data[3].value)



"""Utilities for concurrency optimizations."""

from __future__ import annotations

import asyncio
import functools
import typing

__all__ = ["prevent_concurrency"]

T = typing.TypeVar("T")
AnyCallable = typing.Callable[..., typing.Any]
CallableT = typing.TypeVar("CallableT", bound=AnyCallable)


def prevent_concurrency(func: CallableT) -> CallableT:
    """Prevent function from running concurrently.

    This should be done exclusively for functions that cache their result.
    """

    def wrapper(func: AnyCallable) -> AnyCallable:
        lock: typing.Optional[asyncio.Lock] = None

        @functools.wraps(func)
        async def inner(*args: typing.Any, **kwargs: typing.Any) -> typing.Any:
            nonlocal lock
            if lock is None:
                lock = asyncio.Lock()

            async with lock:
                return await func(*args, **kwargs)

        return inner

    return typing.cast("CallableT", MethodDecorator(func, wrapper))


class MethodDecorator:
    """Descriptor which applies decorators per-instance."""

    method: AnyCallable
    decorator: typing.Callable[[AnyCallable], AnyCallable]
    name: str

    def __init__(
        self,
        method: AnyCallable,
        decorator: typing.Callable[[AnyCallable], AnyCallable],
        *,
        name: typing.Optional[str] = None,
    ) -> None:
        self.method = method  # type: ignore # mypy doesn't understand methods
        self.decorator = decorator  # type: ignore # mypy doesn't understand methods
        self.name = name or self.method.__name__

    def __set_name__(self, owner: type, name: str) -> None:
        self.name = name

    def __get__(self, instance: typing.Optional[T], owner: type[T]) -> AnyCallable:
        if instance is None:
            return self.method

        func = self.decorator(self.method).__get__(instance, type(instance))  # type: ignore # mypy doesn't understand methods
        setattr(instance, self.name, func)
        return func



from __future__ import annotations

import re
import typing

import pydantic

from genshin.models import hoyolab
from genshin.models.model import Aliased, APIModel

from . import abyss, activities
from . import characters as characters_module

__all__ = [
    "AreaExploration",
    "BossKill",
    "Exploration",
    "FullGenshinUserStats",
    "GenshinUserStats",
    "NatlanReputation",
    "NatlanTribe",
    "Offering",
    "PartialGenshinUserStats",
    "Stats",
    "Teapot",
    "TeapotRealm",
]


class ImgTheaterStats(APIModel):
    """Imaginarium Theater stats."""

    unlocked: bool = Aliased("is_unlock")
    max_act: int = Aliased("max_round_id")
    has_data: bool
    has_detail_data: bool


class StygianOnslaughtStats(APIModel):
    """Stygian Onslaught stats."""

    difficulty: int
    name: str
    has_data: bool
    unlocked: bool = Aliased("is_unlock")


# flake8: noqa: E222
class Stats(APIModel):
    """Overall user stats."""

    achievements: int = Aliased("achievement_number")
    days_active: int = Aliased("active_day_number")
    characters: int = Aliased("avatar_number")
    spiral_abyss: str = Aliased("spiral_abyss")
    anemoculi: int = Aliased("anemoculus_number")
    geoculi: int = Aliased("geoculus_number")
    dendroculi: int = Aliased("dendroculus_number")
    electroculi: int = Aliased("electroculus_number")
    hydroculi: int = Aliased("hydroculus_number")
    pyroculi: int = Aliased("pyroculus_number")
    lunoculi: int = Aliased("moonoculus_number")
    common_chests: int = Aliased("common_chest_number")
    exquisite_chests: int = Aliased("exquisite_chest_number")
    precious_chests: int = Aliased("precious_chest_number")
    luxurious_chests: int = Aliased("luxurious_chest_number")
    remarkable_chests: int = Aliased("magic_chest_number")
    unlocked_waypoints: int = Aliased("way_point_number")
    unlocked_domains: int = Aliased("domain_number")
    max_friendship_characters: int = Aliased("full_fetter_avatar_num")

    theater: ImgTheaterStats = Aliased("role_combat")
    stygian: StygianOnslaughtStats = Aliased("hard_challenge")


class Offering(APIModel):
    """Exploration offering."""

    name: str
    level: int
    icon: str = ""


class BossKill(APIModel):
    """Boss kills in exploration"""

    name: str
    kills: int = Aliased("kill_num")


class AreaExploration(APIModel):
    """Area exploration data."""

    name: str
    raw_explored: int = Aliased("exploration_percentage")

    @property
    def explored(self) -> float:
        """The percentage explored. (Note: This can go above 100%)"""
        return self.raw_explored / 10


class NatlanTribe(APIModel):
    """Natlan tribe data."""

    icon: str
    image: str
    name: str
    id: int
    level: int


class NatlanReputation(APIModel):
    """Natlan reputation data."""

    tribes: typing.Sequence[NatlanTribe] = Aliased("tribal_list")


class Exploration(APIModel):
    """Exploration data."""

    id: int
    parent_id: int
    name: str
    raw_explored: int = Aliased("exploration_percentage")

    # deprecated in a sense:
    type: str
    level: int

    icon: str
    inner_icon: str
    background_image: str
    cover: str
    map_url: str

    offerings: typing.Sequence[Offering]
    boss_list: typing.Sequence[BossKill]
    area_exploration_list: typing.Sequence[AreaExploration]
    natlan_reputation: typing.Optional[NatlanReputation] = Aliased("natan_reputation", default=None)

    @property
    def explored(self) -> float:
        """The percentage explored."""
        return self.raw_explored / 10

    @pydantic.field_validator("offerings", mode="before")
    def __add_base_offering(
        cls, offerings: typing.Sequence[typing.Any], info: pydantic.ValidationInfo
    ) -> typing.Sequence[typing.Any]:
        if info.data["type"] == "Reputation" and not any(info.data["type"] == o["name"] for o in offerings):
            offerings = [*offerings, dict(name=info.data["type"], level=info.data["level"])]

        return offerings


class TeapotRealm(APIModel):
    """A specific teapot realm."""

    name: str
    icon: str

    @property
    def id(self) -> int:
        match = re.search(r"\d", self.icon)
        return int(match.group()) if match else 0


class Teapot(APIModel):
    """User's Serenitea Teapot."""

    realms: typing.Sequence[TeapotRealm]
    level: int
    visitors: int = Aliased("visit_num")
    comfort: int = Aliased("comfort_num")
    items: int = Aliased("item_num")
    comfort_name: str = Aliased("comfort_level_name")
    comfort_icon: str = Aliased("comfort_level_icon")


class PartialGenshinUserStats(APIModel):
    """User stats with characters without equipment."""

    info: hoyolab.UserInfo = Aliased("role")
    stats: Stats
    characters: typing.Sequence[characters_module.PartialCharacter] = Aliased("avatars")
    explorations: typing.Sequence[Exploration] = Aliased("world_explorations")
    teapot: typing.Optional[Teapot] = Aliased("homes")

    @pydantic.field_validator("teapot", mode="before")
    def __format_teapot(cls, v: typing.Any) -> typing.Optional[dict[str, typing.Any]]:
        if not v:
            return None
        if isinstance(v, dict):
            return typing.cast("dict[str, typing.Any]", v)
        return {**v[0], "realms": v}

    @pydantic.field_validator("characters", mode="before")
    def __format_characters(cls, v: typing.Sequence[typing.Any]) -> typing.Sequence[typing.Any]:
        return [c for c in v if c["id"] != 0]


class GenshinUserStats(PartialGenshinUserStats):
    """User stats with characters with equipment"""

    characters: typing.Sequence[characters_module.Character] = Aliased("list")


class FullGenshinUserStats(GenshinUserStats):
    """User stats with all data a user can have"""

    abyss: abyss.SpiralAbyssPair
    activities: activities.Activities



import typing
from enum import Enum

import pydantic

from genshin.models.model import APIModel, TZDateTime, prevent_enum_error

__all__ = (
    "ChallengeStatus",
    "ChallengeType",
    "EventWarp",
    "EventWarpCharacter",
    "EventWarpLightCone",
    "HSRBaseEventItem",
    "HSRChallenge",
    "HSREvent",
    "HSREventCalendar",
    "HSREventReward",
    "HSREventStatus",
    "HSREventTimeType",
    "HSREventType",
    "TimeInfo",
)


class HSREventType(Enum):
    """Event type enum."""

    SIGN_IN = "ActivityTypeSign"
    """Daily sign-in event."""
    OTHER = "ActivityTypeOther"
    DOUBLE_REWARDS = "ActivityTypeDouble"
    """Plannar Fissure, Garden of Plenty, etc."""


class HSREventStatus(Enum):
    """Event status enum."""

    OTHER_LOCKED = "OtherActStatusUnopened"
    OTHER_IN_PROGRESS = "OtherActStatusUnFinish"
    DOUBLE_REWARDS_LOCKED = "DoubleActStatusUnopened"
    SIGN_IN_UNCLAIMED = "SignActStatusUnclaimed"

    # Enums below exist but we don't know what they mean for now
    SIGN_STATUS_UNSIGNED_TODAY = "SignStatusUnSignedToday"
    SIGN_STATUS_SIGNED_TODAY = "SignStatusSignedToday"
    SIGN_STATUS_UNCLAIMED = "SignStatusUnclaimed"
    SIGN_STATUS_FINISH = "SignStatusFinish"
    DOUBLE_REWARD_ACT_STATUS_PROGRESS = "DoubleRewardActStatusProgress"
    DOUBLE_REWARD_ACT_STATUS_UNOPENED = "DoubleRewardActStatusUnopened"
    OTHER_ACT_STATUS_FINISH = "OtherActStatusFinish"


class HSREventTimeType(Enum):
    """Event time type enum."""

    LONG = "ActTimeTypeLong"
    """Simulated universe."""
    DEFAULT = "ActTimeTypeDefault"


class ChallengeType(Enum):
    """Challenge type enum."""

    APC_SHADOW = "ChallengeTypeBoss"
    """Apocalyptic shadow."""
    MOC = "ChallengeTypeChasm"
    """Memory of Chaos."""
    PURE_FICTION = "ChallengeTypeStory"


class ChallengeStatus(Enum):
    """Challenge status enum."""

    IN_PROGRESS = "challengeStatusInProgress"
    LOCKED = "challengeStatusUnopened"
    FINISHED = "challengeStatusFinish"


class TimeInfo(APIModel):
    """Time info model."""

    start: TZDateTime = pydantic.Field(alias="start_ts")
    end: TZDateTime = pydantic.Field(alias="end_ts")
    now: TZDateTime


class HSRBaseEvent(APIModel):
    """HSR base event model."""

    time_info: typing.Optional[TimeInfo] = None

    @pydantic.field_validator("time_info", mode="before")
    def __validate_time_info(cls, v: dict[str, typing.Any]) -> typing.Optional[dict[str, typing.Any]]:
        if not v["start_time"]:
            return None
        return v


class HSRBaseEventItem(APIModel):
    """Base event item model."""

    id: int = pydantic.Field(alias="item_id")
    name: str = pydantic.Field(alias="item_name")
    icon: str = pydantic.Field(alias="icon_url")
    path: int = pydantic.Field(alias="avatar_base_type")

    rarity: int
    wiki_url: str

    is_forward: bool  # No clue what this is


class EventWarpCharacter(HSRBaseEventItem):
    """Event warp character model."""

    icon: str = pydantic.Field(alias="icon_url")
    large_icon: str = pydantic.Field(alias="item_avatar_icon_path")
    element: int = pydantic.Field(alias="damage_type")


class EventWarpLightCone(HSRBaseEventItem):
    """Event warp light cone model."""

    icon: str = pydantic.Field(alias="item_url")


class EventWarp(HSRBaseEvent):
    """Event warp model."""

    name: str
    type: str  # Seems to always be 'CardPoolRole'
    characters: typing.Sequence[EventWarpCharacter] = pydantic.Field(alias="avatar_list")
    light_cones: typing.Sequence[EventWarpLightCone] = pydantic.Field(alias="equip_list")

    is_after_version: bool
    """Whether the event happens after last version's update."""
    version: str
    id: int


class HSREventReward(APIModel):
    """HSR event reward model."""

    id: int = pydantic.Field(alias="item_id")
    name: str
    icon: str
    wiki_url: str
    num: int
    rarity: int


class HSREvent(HSRBaseEvent):
    """HSR event model."""

    id: int
    version: str
    name: str
    description: str = pydantic.Field(alias="panel_desc")

    type: typing.Union[HSREventType, str] = pydantic.Field(alias="act_type")
    time_type: typing.Union[HSREventTimeType, str] = pydantic.Field(alias="act_time_type")
    status: typing.Union[HSREventStatus, str] = pydantic.Field(alias="act_status")

    rewards: typing.Sequence[HSREventReward] = pydantic.Field(alias="reward_list")
    total_progress: int
    current_progress: int
    special_reward: typing.Optional[HSREventReward]

    is_after_version: bool
    """Whether the event happens after last version's update."""
    all_finished: bool
    show_text: str

    # No clue what these are
    strategy: str
    multiple_drop_type: int
    multiple_drop_type_list: list[int]
    count_refresh_type: int
    count_value: int
    drop_multiple: int
    panel_id: int
    sort_weight: int

    @pydantic.field_validator("special_reward", mode="after")
    def __validate_special_reward(cls, v: HSREventReward) -> typing.Optional[HSREventReward]:
        if v.id == 0:
            return None
        return v

    @pydantic.field_validator("type", mode="before")
    def __validate_type(cls, v: str) -> typing.Union[HSREventType, str]:
        return prevent_enum_error(v, HSREventType)

    @pydantic.field_validator("time_type", mode="before")
    def __validate_time_type(cls, v: str) -> typing.Union[HSREventTimeType, str]:
        return prevent_enum_error(v, HSREventTimeType)

    @pydantic.field_validator("status", mode="before")
    def __validate_status(cls, v: str) -> typing.Union[HSREventStatus, str]:
        return prevent_enum_error(v, HSREventStatus)

    @pydantic.field_validator("name", mode="after")
    def __format_name(cls, v: str) -> str:
        return v.replace("\\n", " ")

    @pydantic.field_validator("description", mode="after")
    def __format_description(cls, v: str) -> str:
        return v.replace("\\n", "\n")


class HSRChallenge(HSRBaseEvent):
    """HSR challenge model."""

    id: int = pydantic.Field(alias="group_id")
    name: str = pydantic.Field(alias="name_mi18n")

    type: typing.Union[ChallengeType, str] = pydantic.Field(alias="challenge_type")
    status: typing.Union[ChallengeStatus, str]

    rewards: typing.Sequence[HSREventReward] = pydantic.Field(alias="reward_list")
    special_reward: typing.Optional[HSREventReward]
    total_progress: int
    current_progress: int
    show_text: str

    @pydantic.field_validator("special_reward", mode="after")
    def __validate_special_reward(cls, v: typing.Optional[HSREventReward]) -> typing.Optional[HSREventReward]:
        if v is not None and v.id == 0:
            return None
        return v

    @pydantic.field_validator("type", mode="before")
    def __validate_type(cls, v: str) -> typing.Union[ChallengeType, str]:
        return prevent_enum_error(v, ChallengeType)

    @pydantic.field_validator("status", mode="before")
    def __validate_status(cls, v: str) -> typing.Union[ChallengeStatus, str]:
        return prevent_enum_error(v, ChallengeStatus)


class HSREventCalendar(APIModel):
    """HSR event calendar model."""

    character_warps: typing.Sequence[EventWarp] = pydantic.Field(alias="avatar_card_pool_list")
    light_cone_warps: typing.Sequence[EventWarp] = pydantic.Field(alias="equip_card_pool_list")
    events: typing.Sequence[HSREvent] = pydantic.Field(alias="act_list")
    challenges: typing.Sequence[HSRChallenge] = pydantic.Field(alias="challenge_list")

    cur_game_version: str
    """Current game version."""
    now: TZDateTime



"""Battle chronicle models."""

from .abyss import *
from .act_calendar import *
from .activities import *
from .char_master import *
from .characters import *
from .hard_challenge import *
from .img_theater import *
from .notes import *
from .stats import *
from .tcg import *



"""Genshin diary models."""

import enum
import typing

from genshin.models.model import Aliased, APIModel, TZDateTime

__all__ = [
    "BaseDiary",
    "DayDiaryData",
    "Diary",
    "DiaryAction",
    "DiaryActionCategory",
    "DiaryPage",
    "DiaryType",
    "MonthDiaryData",
    "StarRailDayDiaryData",
    "StarRailDiary",
    "StarRailDiaryAction",
    "StarRailDiaryActionCategory",
    "StarRailDiaryPage",
    "StarRailDiaryType",
    "StarRailMonthDiaryData",
]


class DiaryType(enum.IntEnum):
    """Type of diary pages."""

    PRIMOGEMS = 1
    """Primogems."""

    MORA = 2
    """Mora."""


class BaseDiary(APIModel):
    """Base model for diary and diary page."""

    uid: int
    server: str = Aliased("region")
    nickname: str = ""
    month: int = Aliased("data_month")


class DiaryActionCategory(APIModel):
    """Diary category for primogems."""

    id: int = Aliased("action_id")
    name: str = Aliased("action")
    amount: int = Aliased("num")
    percentage: int = Aliased("percent")


class MonthDiaryData(APIModel):
    """Diary data for a month."""

    current_primogems: int
    current_mora: int
    last_primogems: int
    last_mora: int
    primogems_rate: int = Aliased("primogem_rate")
    mora_rate: int
    categories: typing.Sequence[DiaryActionCategory] = Aliased("group_by")


class DayDiaryData(APIModel):
    """Diary data for a day."""

    current_primogems: int
    current_mora: int


class Diary(BaseDiary):
    """Traveler's diary."""

    data: MonthDiaryData = Aliased("month_data")
    day_data: DayDiaryData

    @property
    def month_data(self) -> MonthDiaryData:
        return self.data


class DiaryAction(APIModel):
    """Action which earned currency."""

    action_id: int
    action: str
    time: TZDateTime
    amount: int = Aliased("num")


class DiaryPage(BaseDiary):
    """Page of a diary."""

    actions: typing.Sequence[DiaryAction] = Aliased("list")


class StarRailDiaryActionCategory(APIModel):
    """Diary category for rails_pass ."""

    id: str = Aliased("action")
    name: str = Aliased("action_name")
    amount: int = Aliased("num")
    percentage: int = Aliased("percent")


class StarRailMonthDiaryData(APIModel):
    """Diary data for a month."""

    current_hcoin: int
    current_rails_pass: int
    last_hcoin: int
    last_rails_pass: int
    hcoin_rate: int
    rails_rate: int
    categories: typing.Sequence[StarRailDiaryActionCategory] = Aliased("group_by")


class StarRailDayDiaryData(APIModel):
    """Diary data for a day."""

    current_hcoin: int
    current_rails_pass: int
    last_hcoin: int
    last_rails_pass: int


class StarRailDiary(BaseDiary):
    """Traveler's diary."""

    data: StarRailMonthDiaryData = Aliased("month_data")
    day_data: StarRailDayDiaryData

    @property
    def month_data(self) -> StarRailMonthDiaryData:
        return self.data


class StarRailDiaryType(enum.IntEnum):
    """Type of diary pages."""

    STELLARJADE = 1
    """STELLARJADE."""

    STARRAILPASS = 2
    """STARRAILPASS."""


class StarRailDiaryAction(APIModel):
    """Action which earned currency."""

    action: str
    action_name: str
    time: TZDateTime
    amount: int = Aliased("num")


class StarRailDiaryPage(BaseDiary):
    """Page of a diary."""

    actions: typing.Sequence[StarRailDiaryAction] = Aliased("list")



"""Dynamic secret generation."""

import hashlib
import json
import random
import string
import time
import typing

from genshin import constants, types

__all__ = [
    "generate_app_login_ds",
    "generate_cn_dynamic_secret",
    "generate_dynamic_secret",
    "generate_geetest_ds",
    "generate_passport_ds",
    "get_ds_headers",
]


def generate_dynamic_secret(salt: str = constants.DS_SALT[types.Region.OVERSEAS]) -> str:
    """Create a new overseas dynamic secret."""
    t = int(time.time())
    r = "".join(random.choices(string.ascii_letters, k=6))
    h = hashlib.md5(f"salt={salt}&t={t}&r={r}".encode()).hexdigest()
    return f"{t},{r},{h}"


def generate_app_login_ds(body: dict[str, typing.Any]) -> str:
    """Create a dynamic secret for app login."""
    # Follows exactly the same implementation as in RequestUtils.createSign()
    # of the official app. See comment in auth.py on how to decompile the app.
    t = int(time.time())
    r = "".join(random.choices(string.ascii_letters + string.digits, k=6))
    b = json.dumps(body, separators=(",", ":"))
    h = hashlib.md5(f"salt={constants.DS_SALT['app_login']}&t={t}&r={r}&b={b}&q=".encode()).hexdigest()
    return f"{t},{r},{h}"


def generate_cn_dynamic_secret(
    body: typing.Any = None,
    query: typing.Optional[typing.Mapping[str, typing.Any]] = None,
    *,
    salt: str = constants.DS_SALT[types.Region.CHINESE],
) -> str:
    """Create a new chinese dynamic secret."""
    t = int(time.time())
    r = random.randint(100001, 200000)
    b = json.dumps(body) if body else ""
    q = "&".join(f"{k}={v}" for k, v in sorted(query.items())) if query else ""

    h = hashlib.md5(f"salt={salt}&t={t}&r={r}&b={b}&q={q}".encode()).hexdigest()
    return f"{t},{r},{h}"


def get_ds_headers(
    region: types.Region,
    data: typing.Any = None,
    params: typing.Optional[typing.Mapping[str, typing.Any]] = None,
    lang: typing.Optional[str] = None,
) -> dict[str, typing.Any]:
    """Get ds http headers."""
    if region == types.Region.OVERSEAS:
        ds_headers = {
            "x-rpc-app_version": "1.5.0",
            "x-rpc-client_type": "5",
            "x-rpc-language": lang,
            "x-rpc-lang": lang,
            "ds": generate_dynamic_secret(),
        }
    elif region == types.Region.CHINESE:
        ds_headers = {
            "x-rpc-app_version": "2.11.1",
            "x-rpc-client_type": "5",
            "ds": generate_cn_dynamic_secret(data, params),
        }
    else:
        raise TypeError(f"{region!r} is not a valid region.")
    return ds_headers


def generate_passport_ds(body: typing.Mapping[str, typing.Any]) -> str:
    """Create a dynamic secret for Miyoushe passport API."""
    salt = constants.DS_SALT["cn_passport"]
    t = int(time.time())
    r = "".join(random.sample(string.ascii_letters, 6))
    b = json.dumps(body)
    h = hashlib.md5(f"salt={salt}&t={t}&r={r}&b={b}&q=".encode()).hexdigest()
    result = f"{t},{r},{h}"
    return result


def generate_geetest_ds(region: types.Region) -> str:
    """Create a dynamic secret for geetest API endpoint."""
    t = int(time.time())
    r = random.randint(100000, 200000)
    h = hashlib.md5(f"salt={constants.DS_SALT[region]}&t={t}&r={r}&b=&q=is_high=false".encode()).hexdigest()
    return f"{t},{r},{h}"
